"""Module to support importing the datasets used by the documentation."""

from typing import Any, cast
from pathlib import Path

from pydicom.data import get_testdata_file
from pydicom.filereader import dcmread


# All datasets included here must be available in the package itself
#   NOT via the pydicom-data download method
_DATASETS: dict[str, str] = {
    "ct": cast(str, get_testdata_file("CT_small.dcm")),
    "dicomdir": cast(str, get_testdata_file("DICOMDIR")),
    "jpeg2k": cast(str, get_testdata_file("examples_jpeg2k.dcm")),
    "mr": cast(str, get_testdata_file("MR_small.dcm")),
    "no_meta": cast(str, get_testdata_file("no_meta.dcm")),
    "overlay": cast(str, get_testdata_file("examples_overlay.dcm")),
    "palette_color": cast(str, get_testdata_file("examples_palette.dcm")),
    "rgb_color": cast(str, get_testdata_file("examples_rgb_color.dcm")),
    "rt_dose": cast(str, get_testdata_file("rtdose.dcm")),
    "rt_plan": cast(str, get_testdata_file("rtplan.dcm")),
    "rt_ss": cast(str, get_testdata_file("rtstruct.dcm")),
    "waveform": cast(str, get_testdata_file("waveform_ecg.dcm")),
    "ybr_color": cast(str, get_testdata_file("examples_ybr_color.dcm")),
}


def _get_path(name: str) -> Path:
    """Return the path to the example dataset with the attribute name `name` as
    :class:`pathlib.Path`.
    """
    if name in _DATASETS:
        return Path(_DATASETS[name])

    raise ValueError(f"No example dataset exists with the name '{name}'")


def __getattr__(name: str) -> Any:
    """Return module level attributes."""
    if name in _DATASETS:
        return dcmread(_DATASETS[name], force=True)

    if name == "get_path":
        return _get_path

    raise AttributeError(f"module '{__name__}' has no attribute '{name}'")



# Copyright 2008-2018 pydicom authors. See LICENSE file for details.
# TODO: remove module in v4.0
"""Utility functions used in the pixel data handlers."""

from sys import byteorder
from typing import Any

try:
    import numpy as np
except ImportError:
    pass

from pydicom import config
from pydicom.misc import warn_and_log


from pydicom.pixels.processing import (
    apply_color_lut as _apply_color_lut,
    apply_modality_lut as _apply_modality_lut,
    apply_voi_lut as _apply_voi_lut,
    apply_voi as _apply_voi,
    apply_windowing as _apply_windowing,
    convert_color_space as _convert_color_space,
)
from pydicom.pixels.utils import (
    expand_ybr422 as _expand_ybr422,
    get_expected_length as _get_expected_length,
    get_image_pixel_ids as _get_image_pixel_ids,
    get_j2k_parameters as _get_j2k_parameters,
    get_nr_frames as _get_nr_frames,
    pack_bits as _pack_bits,
    pixel_dtype as _pixel_dtype,
    reshape_pixel_array as _reshape_pixel_array,
    unpack_bits as _unpack_bits,
)


def _dtype_corrected_for_endianness(
    is_little_endian: bool, numpy_dtype: "np.dtype"
) -> "np.dtype":
    """Return a :class:`numpy.dtype` corrected for system and :class:`Dataset`
    endianness.

    .. deprecated:: 3.0

        This function will be removed in v4.0.

    Parameters
    ----------
    is_little_endian : bool
        The endianness of the affected :class:`~pydicom.dataset.Dataset`.
    numpy_dtype : numpy.dtype
        The numpy data type used for the *Pixel Data* without considering
        endianness.

    Raises
    ------
    ValueError
        If `is_little_endian` is ``None``, e.g. not initialized.

    Returns
    -------
    numpy.dtype
        The numpy data type used for the *Pixel Data* without considering
        endianness.
    """
    if is_little_endian is None:
        raise ValueError(
            "Dataset attribute 'is_little_endian' "
            "has to be set before writing the dataset"
        )

    if is_little_endian != (byteorder == "little"):
        return numpy_dtype.newbyteorder("S")

    return numpy_dtype


_DEPRECATED = {
    "apply_color_lut": _apply_color_lut,
    "apply_modality_lut": _apply_modality_lut,
    "apply_voi_lut": _apply_voi_lut,
    "apply_voi": _apply_voi,
    "apply_windowing": _apply_windowing,
    "convert_color_space": _convert_color_space,
    "pack_bits": _pack_bits,
    "unpack_bits": _unpack_bits,
}
_DEPRECATED_UTIL = {
    "expand_ybr422": _expand_ybr422,
    "get_expected_length": _get_expected_length,
    "get_image_pixel_ids": _get_image_pixel_ids,
    "get_j2k_parameters": _get_j2k_parameters,
    "get_nr_frames": _get_nr_frames,
    "pixel_dtype": _pixel_dtype,
    "reshape_pixel_array": _reshape_pixel_array,
}


def __getattr__(name: str) -> Any:
    if name in _DEPRECATED and not config._use_future:
        msg = (
            "The 'pydicom.pixel_data_handlers' module will be removed "
            f"in v4.0, please use 'from pydicom.pixels import {name}' instead"
        )
        warn_and_log(msg, DeprecationWarning)
        return _DEPRECATED[name]

    if name in _DEPRECATED_UTIL and not config._use_future:
        msg = (
            "The 'pydicom.pixel_data_handlers' module will be removed "
            f"in v4.0, please use 'from pydicom.pixels.utils import {name}' instead"
        )
        warn_and_log(msg, DeprecationWarning)
        return _DEPRECATED_UTIL[name]

    if name == "dtype_corrected_for_endianness" and not config._use_future:
        msg = (
            "'dtype_corrected_for_endianness' is deprecated and will be "
            "removed in v4.0"
        )
        warn_and_log(msg, DeprecationWarning)
        return _dtype_corrected_for_endianness

    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



# Copyright 2008-2024 pydicom authors. See LICENSE file for details.
"""Interface for *Pixel Data* encoding, not intended to be used directly."""

from typing import cast

from pydicom.pixels.encoders.base import EncodeRunner
from pydicom.uid import RLELossless

try:
    import gdcm

    GDCM_VERSION = tuple(int(x) for x in gdcm.Version.GetVersion().split("."))
    HAVE_GDCM = True
except ImportError:
    HAVE_GDCM = False


ENCODER_DEPENDENCIES = {
    RLELossless: ("gdcm>=3.0.10",),
}


def is_available(uid: str) -> bool:
    """Return ``True`` if a pixel data encoder for `uid` is available for use,
    ``False`` otherwise.
    """
    if not HAVE_GDCM or GDCM_VERSION < (3, 0, 10):
        return False

    return uid in ENCODER_DEPENDENCIES


def encode_pixel_data(src: bytes, runner: EncodeRunner) -> bytes:
    """Return the encoded image data in `src`.

    Parameters
    ----------
    src : bytes
        The raw image frame data to be encoded.
    runner : pydicom.pixels.encoders.base.EncodeRunner
        The runner managing the encoding process.

    Returns
    -------
    bytes
        The encoded image data.
    """
    byteorder = runner.get_option("byteorder", "<")
    if byteorder == ">":
        raise ValueError("Unsupported option \"byteorder = '>'\"")

    return _ENCODERS[runner.transfer_syntax](src, runner)


def _rle_encode(src: bytes, runner: EncodeRunner) -> bytes:
    """Return RLE encoded image data from `src`.

    Parameters
    ----------
    src : bytes
        The raw image frame data to be encoded.
    runner : pydicom.pixels.encoders.base.EncodeRunner
        The runner managing the encoding process.

    Returns
    -------
    bytes
        The encoded image data.
    """
    if runner.bits_allocated > 32:
        raise ValueError("Unable to encode more than 32-bit data")

    # Create a gdcm.Image with the uncompressed `src` data
    pi = gdcm.PhotometricInterpretation.GetPIType(runner.photometric_interpretation)

    # GDCM's null photometric interpretation gets used for invalid values
    if pi == gdcm.PhotometricInterpretation.PI_END:
        raise ValueError(
            f"Invalid photometric interpretation '{runner.photometric_interpretation}'"
        )

    # `src` uses little-endian byte ordering
    ts = gdcm.TransferSyntax.ImplicitVRLittleEndian

    # Must use ImageWriter().GetImage() to create a gdcmImage
    #   also have to make sure `writer` doesn't go out of scope
    writer = gdcm.ImageWriter()
    image = writer.GetImage()
    image.SetNumberOfDimensions(2)
    image.SetDimensions((runner.columns, runner.rows, 1))
    image.SetPhotometricInterpretation(gdcm.PhotometricInterpretation(pi))
    image.SetTransferSyntax(gdcm.TransferSyntax(ts))

    pixel_format = gdcm.PixelFormat(
        runner.samples_per_pixel,
        runner.bits_allocated,
        runner.bits_stored,
        runner.bits_stored - 1,
        runner.pixel_representation,
    )
    image.SetPixelFormat(pixel_format)
    if runner.samples_per_pixel > 1:
        # Default `src` is planar configuration 0 (i.e. R1 G1 B1 R2 G2 B2)
        image.SetPlanarConfiguration(0)

    # Add the Pixel Data element and set the value to `src`
    elem = gdcm.DataElement(gdcm.Tag(0x7FE0, 0x0010))
    elem.SetByteStringValue(src)
    image.SetDataElement(elem)

    # Converts an image to match the set transfer syntax
    converter = gdcm.ImageChangeTransferSyntax()

    # Set up the converter with the intended transfer syntax...
    rle = gdcm.TransferSyntax.GetTSType(runner.transfer_syntax)
    converter.SetTransferSyntax(gdcm.TransferSyntax(rle))
    # ...and image to be converted
    converter.SetInput(image)

    # Perform the conversion, returns bool
    # 'PALETTE COLOR' and a lossy transfer syntax will return False
    result = converter.Change()
    if not result:
        raise RuntimeError(
            "ImageChangeTransferSyntax.Change() returned a failure result"
        )

    # A new gdcmImage with the converted pixel data element
    image = converter.GetOutput()

    # The element's value is the encapsulated encoded pixel data
    seq = image.GetDataElement().GetSequenceOfFragments()

    # RLECodec::Code() uses only 1 fragment per frame
    if seq is None or seq.GetNumberOfFragments() != 1:
        # Covers both no sequence and unexpected number of fragments
        raise RuntimeError("Unexpected number of fragments found in the 'Pixel Data'")

    fragment = seq.GetFragment(0).GetByteValue().GetBuffer()
    return cast(bytes, fragment.encode("utf-8", "surrogateescape"))


_ENCODERS = {RLELossless: _rle_encode}



# Copyright 2008-2024 pydicom authors. See LICENSE file for details.
"""Use Python to decode RLE Lossless encoded *Pixel Data*.

This module is not intended to be used directly.
"""

from struct import unpack

from pydicom.misc import warn_and_log
from pydicom.pixels.decoders.base import DecodeRunner
from pydicom.uid import RLELossless


DECODER_DEPENDENCIES = {RLELossless: ()}


def is_available(uid: str) -> bool:
    """Return ``True`` if a pixel data decoder for `uid` is available for use,
    ``False`` otherwise.
    """
    return uid in DECODER_DEPENDENCIES


def _decode_frame(src: bytes, runner: DecodeRunner) -> bytearray:
    """Wrapper for use with the decoder interface.

    Parameters
    ----------
    src : bytes
        A single frame of RLE encoded data.
    runner : pydicom.pixels.decoders.base.DecodeRunner


        Required parameters:

        * `rows`: int
        * `columns`: int
        * `samples_per_pixel`: int
        * `bits_allocated`: int

        Optional parameters:

        * `rle_segment_order`: str, "<" for little endian segment order, or
          ">" for big endian (default)

    Returns
    -------
    bytearray
        The decoded frame, ordered as planar configuration 1.
    """
    frame = _rle_decode_frame(
        src,
        runner.rows,
        runner.columns,
        runner.samples_per_pixel,
        runner.bits_allocated,
        runner.get_option("rle_segment_order", ">"),
    )
    # Update the runner options to ensure the reshaping is correct
    # Only do this if we successfully decoded the frame
    runner.set_option("planar_configuration", 1)

    return frame


def _rle_decode_frame(
    src: bytes,
    rows: int,
    columns: int,
    nr_samples: int,
    nr_bits: int,
    segment_order: str = ">",
) -> bytearray:
    """Decodes a single frame of RLE encoded data.

    Each frame may contain up to 15 segments of encoded data.

    Parameters
    ----------
    src : bytes
        The RLE frame data
    rows : int
        The number of output rows
    columns : int
        The number of output columns
    nr_samples : int
        Number of samples per pixel (e.g. 3 for RGB data).
    nr_bits : int
        Number of bits per sample - must be a multiple of 8
    segment_order : str
        The segment order of the `data`, '>' for big endian (default),
        '<' for little endian (non-conformant).

    Returns
    -------
    bytearray
        The frame's decoded data in little endian and planar configuration 1
        byte ordering (i.e. for RGB data this is all red pixels then all
        green then all blue, with the bytes for each pixel ordered from
        MSB to LSB when reading left to right).
    """
    if nr_bits % 8:
        raise NotImplementedError(
            f"Unable to decode RLE encoded pixel data with {nr_bits} bits allocated"
        )

    # Parse the RLE Header
    offsets = _rle_parse_header(src[:64])
    nr_segments = len(offsets)

    # Check that the actual number of segments is as expected
    bytes_per_sample = nr_bits // 8
    if nr_segments != nr_samples * bytes_per_sample:
        raise ValueError(
            "The number of RLE segments in the pixel data doesn't match the "
            f"expected amount ({nr_segments} vs. {nr_samples * bytes_per_sample} "
            "segments)"
        )

    # Ensure the last segment gets decoded
    offsets.append(len(src))

    # Preallocate with null bytes
    decoded = bytearray(rows * columns * nr_samples * bytes_per_sample)

    # Example:
    # RLE encoded data is ordered like this (for 16-bit, 3 sample):
    #  Segment: 0     | 1     | 2     | 3     | 4     | 5
    #           R MSB | R LSB | G MSB | G LSB | B MSB | B LSB
    #  A segment contains only the MSB or LSB parts of all the sample pixels

    # To minimise the amount of array manipulation later, and to make things
    # faster we interleave each segment in a manner consistent with a planar
    # configuration of 1 (and use little endian byte ordering):
    #    All red samples             | All green samples           | All blue
    #    Pxl 1   Pxl 2   ... Pxl N   | Pxl 1   Pxl 2   ... Pxl N   | ...
    #    LSB MSB LSB MSB ... LSB MSB | LSB MSB LSB MSB ... LSB MSB | ...

    # `stride` is the total number of bytes of each sample plane
    stride = bytes_per_sample * rows * columns
    for sample_number in range(nr_samples):
        le_gen = range(bytes_per_sample)
        byte_offsets = le_gen if segment_order == "<" else reversed(le_gen)
        for byte_offset in byte_offsets:
            # Decode the segment
            ii = sample_number * bytes_per_sample + byte_offset
            # ii is 1, 0, 3, 2, 5, 4 for the example above
            # This is where the segment order correction occurs
            segment = _rle_decode_segment(src[offsets[ii] : offsets[ii + 1]])

            # Check that the number of decoded bytes is correct
            actual_length = len(segment)
            if actual_length < rows * columns:
                raise ValueError(
                    "The amount of decoded RLE segment data doesn't match the "
                    f"expected amount ({actual_length} vs. {rows * columns} bytes)"
                )
            elif actual_length != rows * columns:
                warn_and_log(
                    "The decoded RLE segment contains non-conformant padding "
                    f"- {actual_length} vs. {rows * columns} bytes expected"
                )

            if segment_order == ">":
                byte_offset = bytes_per_sample - byte_offset - 1

            # For 100 pixel/plane, 32-bit, 3 sample data, `start` will be
            #   0, 1, 2, 3, 400, 401, 402, 403, 800, 801, 802, 803
            start = byte_offset + (sample_number * stride)
            decoded[start : start + stride : bytes_per_sample] = segment[
                : rows * columns
            ]

    return decoded


def _rle_decode_segment(src: bytes) -> bytearray:
    """Return a single segment of decoded RLE data as bytearray.

    Parameters
    ----------
    buffer : bytes
        The segment data to be decoded.

    Returns
    -------
    bytearray
        The decoded segment.
    """
    result = bytearray()
    pos = 0
    result_extend = result.extend

    try:
        while True:
            # header_byte is N + 1
            header_byte = src[pos] + 1
            pos += 1
            if header_byte > 129:
                # Extend by copying the next byte (-N + 1) times
                # however since using uint8 instead of int8 this will be
                # (256 - N + 1) times
                result_extend(src[pos : pos + 1] * (258 - header_byte))
                pos += 1
            elif header_byte < 129:
                # Extend by literally copying the next (N + 1) bytes
                result_extend(src[pos : pos + header_byte])
                pos += header_byte

    except IndexError:
        pass

    return result


def _rle_parse_header(header: bytes) -> list[int]:
    """Return a list of byte offsets for the segments in RLE data.

    **RLE Header Format**

    The RLE Header contains the number of segments for the image and the
    starting offset of each segment. Each of these numbers is represented as
    an unsigned long stored in little-endian. The RLE Header is 16 long words
    in length (i.e. 64 bytes) which allows it to describe a compressed image
    with up to 15 segments. All unused segment offsets shall be set to zero.

    As an example, the table below describes an RLE Header with 3 segments as
    would typically be used with 8-bit RGB or YCbCr data (with 1 segment per
    channel).

    +--------------+---------------------------------+------------+
    | Byte  offset | Description                     | Value      |
    +==============+=================================+============+
    | 0            | Number of segments              | 3          |
    +--------------+---------------------------------+------------+
    | 4            | Offset of segment 1, N bytes    | 64         |
    +--------------+---------------------------------+------------+
    | 8            | Offset of segment 2, M bytes    | 64 + N     |
    +--------------+---------------------------------+------------+
    | 12           | Offset of segment 3             | 64 + N + M |
    +--------------+---------------------------------+------------+
    | 16           | Offset of segment 4 (not used)  | 0          |
    +--------------+---------------------------------+------------+
    | ...          | ...                             | 0          |
    +--------------+---------------------------------+------------+
    | 60           | Offset of segment 15 (not used) | 0          |
    +--------------+---------------------------------+------------+

    Parameters
    ----------
    header : bytes
        The RLE header data (i.e. the first 64 bytes of an RLE frame).

    Returns
    -------
    list of int
        The byte offsets for each segment in the RLE data.

    Raises
    ------
    ValueError
        If there are more than 15 segments or if the header is not 64 bytes
        long.

    References
    ----------
    DICOM Standard, Part 5, :dcm:`Annex G<part05/chapter_G.html>`
    """
    if len(header) != 64:
        raise ValueError("The RLE header can only be 64 bytes long")

    nr_segments = unpack("<L", header[:4])[0]
    if nr_segments > 15:
        raise ValueError(
            f"The RLE header specifies an invalid number of segments ({nr_segments})"
        )

    return list(unpack(f"<{nr_segments}L", header[4 : 4 * (nr_segments + 1)]))



# Copyright 2008-2024 pydicom authors. See LICENSE file for details.
"""Access code dictionary information"""

from itertools import chain
import inspect
from typing import cast, Any
from collections.abc import KeysView, Iterable

from pydicom.sr.coding import Code
from pydicom.sr._concepts_dict import concepts as CONCEPTS
from pydicom.sr._cid_dict import name_for_cid, cid_concepts as CID_CONCEPTS


# Reverse lookup for cid names
cid_for_name = {v: k for k, v in name_for_cid.items()}


def _filtered(source: Iterable[str], filters: Iterable[str]) -> list[str]:
    """Return a sorted list of filtered str.

    Parameters
    ----------
    source : Iterable[str]
        The iterable of str to be filtered.
    filters : Iterable[str]
        An iterable containing patterns for which values are to be included
        in the results.

    Returns
    -------
    List[str]
        A sorted list of unique values from `source`, filtered by including
        case-insensitive partial or full matches against the values
        in `filters`.
    """
    if not filters:
        return sorted(set(source))

    filters = [f.lower() for f in filters]

    return sorted(
        set(val for val in source if any((f in val.lower()) for f in filters))
    )


CIDValueType = dict[str, tuple[str, list[int]]]
ConceptsType = dict[str, CIDValueType]
SnomedMappingType = dict[str, dict[str, str]]


class Collection:
    """Interface for a collection of concepts, such as SNOMED-CT, or a DICOM CID.

    .. versionadded:: 3.0
    """

    repr_format = "{} = {}"

    def __init__(self, name: str) -> None:
        """Create a new collection.

        Parameters
        ----------
        name : str
            The name of the collection, should either be a key in the
            ``sr._concepts_dict.concepts`` :class:`dict` or a CID name for
            a CID in ``sr._cid_dict.cid_concepts`` such as ``"CID1234"``.
        """
        if not name.upper().startswith("CID"):
            self._name = name
            # dict[str, dict[str, tuple(str, list[int])]]
            # {'ACEInhibitor': {'41549009': ('ACE inhibitor', [3760])},
            self._scheme_data = CONCEPTS[name]
        else:
            self._name = f"CID{name[3:]}"
            # dict[str, list[str]]
            # {'SCT': ['Pericardium', 'Pleura', 'LeftPleura', 'RightPleura']}
            self._cid_data = CID_CONCEPTS[int(name[3:])]

        self._concepts: dict[str, Code] = {}

    @property
    def concepts(self) -> dict[str, Code]:
        """Return a :class:`dict` of {SR identifiers: codes}"""
        if not self._concepts:
            self._concepts = {name: getattr(self, name) for name in self.dir()}

        return self._concepts

    def __contains__(self, item: str | Code) -> bool:
        """Checks whether a given code is a member of the collection.

        Parameters
        ----------
        item : pydicom.sr.coding.Code | str
            The code to check for as either the code or the corresponding
            keyword.

        Returns
        -------
        bool
            Whether the collection contains the `code`
        """
        if isinstance(item, str):
            try:
                code = getattr(self, item)
            except AttributeError:
                return False
        else:
            code = item

        return code in self.concepts.values()

    def __dir__(self) -> list[str]:
        """Return a list of available concept keywords.

        List of attributes is used, for example, in auto-completion in editors
        or command-line environments.
        """
        meths = {v[0] for v in inspect.getmembers(type(self), inspect.isroutine)}
        props = {v[0] for v in inspect.getmembers(type(self), inspect.isdatadescriptor)}
        sr_names = set(self.dir())

        return sorted(props | meths | sr_names)

    def dir(self, *filters: str) -> list[str]:
        """Return an sorted list of concept keywords based on a partial match.

        Parameters
        ----------
        filters : str
            Zero or more string arguments to the function. Used for
            case-insensitive match to any part of the SR keyword.

        Returns
        -------
        list of str
            The matching keywords. If no `filters` are used then all
            keywords are returned.
        """
        # CID_CONCEPTS: Dict[int, Dict[str, List[str]]]
        if self.is_cid:
            return _filtered(chain.from_iterable(self._cid_data.values()), filters)

        return _filtered(self._scheme_data, filters)

    def __getattr__(self, name: str) -> Code:
        """Return the :class:`~pydicom.sr.Code` corresponding to `name`.

        Parameters
        ----------
        name : str
            A camel case version of the concept's code meaning, such as
            ``"FontanelOfSkull" in the SCT coding scheme.

        Returns
        -------
        pydicom.sr.Code
            The :class:`~pydicom.sr.Code` corresponding to `name`.
        """
        if self.name.startswith("CID"):
            # Try DICOM's CID collections
            matches = [
                scheme
                for scheme, keywords in self._cid_data.items()
                if name in keywords
            ]
            if not matches:
                raise AttributeError(
                    f"No matching code for keyword '{name}' in {self.name}"
                )

            if len(matches) > 1:
                # Shouldn't happen, but just in case
                raise RuntimeError(
                    f"Multiple schemes found to contain the keyword '{name}' in "
                    f"{self.name}: {', '.join(matches)}"
                )

            scheme = matches[0]
            identifiers = cast(CIDValueType, CONCEPTS[scheme][name])

            if len(identifiers) == 1:
                code, val = list(identifiers.items())[0]
            else:
                cid = int(self.name[3:])
                _matches = [
                    (code, val) for code, val in identifiers.items() if cid in val[1]
                ]
                if len(_matches) > 1:
                    # Shouldn't happen, but just in case
                    codes = ", ".join(v[0] for v in _matches)
                    raise RuntimeError(
                        f"Multiple codes found for keyword '{name}' in {self.name}: "
                        f"{codes}"
                    )

                code, val = _matches[0]

            return Code(value=code, meaning=val[0], scheme_designator=scheme)

        # Try concept collections such as SCT, DCM, etc
        try:
            entries = cast(CIDValueType, self._scheme_data[name])
        except KeyError:
            raise AttributeError(
                f"No matching code for keyword '{name}' in scheme '{self.name}'"
            )

        if len(entries) > 1:
            # val is {"code": ("meaning", [cid1, cid2, ...], "code": ...}
            code_values = ", ".join(entries.keys())
            raise RuntimeError(
                f"Multiple codes found for keyword '{name}' in scheme '{self.name}': "
                f"{code_values}"
            )

        code = list(entries.keys())[0]  # get first and only
        meaning, cids = entries[code]

        return Code(value=code, meaning=meaning, scheme_designator=self.name)

    @property
    def is_cid(self) -> bool:
        """Return ``True`` if the collection is one of the DICOM CIDs"""
        return self.name.startswith("CID")

    @property
    def name(self) -> str:
        """Return the name of the collection."""
        return self._name

    def __repr__(self) -> str:
        """Return a representation of the collection."""
        concepts = [
            self.repr_format.format(name, concept)
            for name, concept in self.concepts.items()
        ]

        return f"{self.name}\n" + "\n".join(concepts)

    @property
    def scheme_designator(self) -> str:
        """Return the scheme designator for the collection."""
        return self.name

    def __str__(self) -> str:
        """Return a string representation of the collection."""
        len_names = max(len(n) for n in self.concepts.keys()) + 2
        len_codes = max(len(c[0]) for c in self.concepts.values()) + 2
        len_schemes = max(len(c[1]) for c in self.concepts.values()) + 2

        # Ensure each column is at least X characters wide
        len_names = max(len_names, 11)
        len_codes = max(len_codes, 6)
        len_schemes = max(len_schemes, 8)

        if self.is_cid:
            fmt = f"{{:{len_names}}} {{:{len_codes}}} {{:{len_schemes}}} {{}}"

            s = [self.name]
            s.append(fmt.format("Attribute", "Code", "Scheme", "Meaning"))
            s.append(fmt.format("---------", "----", "------", "-------"))
            s.append(
                "\n".join(
                    fmt.format(name, *concept)
                    for name, concept in self.concepts.items()
                )
            )
        else:
            fmt = f"{{:{len_names}}} {{:{len_codes}}} {{}}"

            s = [f"Scheme: {self.name}"]
            s.append(fmt.format("Attribute", "Code", "Meaning"))
            s.append(fmt.format("---------", "----", "-------"))

            s.append(
                "\n".join(
                    fmt.format(name, concept[0], concept[2])
                    for name, concept in self.concepts.items()
                )
            )

        return "\n".join(s)

    def trait_names(self) -> list[str]:
        """Return a list of valid names for auto-completion code.

        Used in IPython, so that data element names can be found and offered
        for autocompletion on the IPython command line.
        """
        return dir(self)


class Concepts:
    """Management class for the available concept collections.

    .. versionadded:: 3.0
    """

    def __init__(self, collections: list[Collection]) -> None:
        """Create a new concepts management class instance.

        Parameters
        ----------
        collections : list[Collection]
            A list of the available concept collections.
        """
        self._collections = {c.name: c for c in collections}

    @property
    def collections(self) -> KeysView[str]:
        """Return the names of the available concept collections."""
        return self._collections.keys()

    def __getattr__(self, name: str) -> Any:
        """Return the concept collection corresponding to `name`.

        Parameters
        ----------
        name : str
            The scheme designator or CID name for the collection to be returned.
        """
        if name.upper().startswith("CID"):
            name = f"CID{name[3:]}"

        if name in self._collections:
            return self._collections[name]

        raise AttributeError(
            f"'{type(self).__name__}' object has no attribute '{name}'"
        )

    def schemes(self) -> list[str]:
        """Return a list of available scheme designations."""
        return [c for c in self._collections.keys() if not c.startswith("CID")]

    def CIDs(self) -> list[str]:
        """Return a list of available CID names."""
        return [c for c in self._collections.keys() if c.startswith("CID")]


# Named concept collections like SNOMED-CT, etc
_collections = [Collection(designator) for designator in CONCEPTS]
# DICOM CIDs
_collections.extend(Collection(f"CID{cid}") for cid in name_for_cid)

codes = Concepts(_collections)



# Copyright 2008-2024 pydicom authors. See LICENSE file for details.

from pydicom.pixels.decoders.base import get_decoder
from pydicom.pixels.encoders.base import get_encoder
from pydicom.pixels.processing import (
    apply_color_lut,
    apply_icc_profile,
    apply_modality_lut,
    apply_presentation_lut,
    apply_rescale,
    apply_voi_lut,
    apply_voi,
    apply_windowing,
    convert_color_space,
    create_icc_transform,
)
from pydicom.pixels.utils import (
    as_pixel_options,
    compress,
    decompress,
    iter_pixels,
    pack_bits,
    pixel_array,
    set_pixel_data,
    unpack_bits,
)



# Copyright 2008-2018 pydicom authors. See LICENSE file for details.
"""Use the `pillow <https://python-pillow.org/>`_ Python package
to decode *Pixel Data*.
"""

import io
import logging
from typing import TYPE_CHECKING, cast

if TYPE_CHECKING:  # pragma: no cover
    from pydicom.dataset import Dataset

try:
    import numpy

    HAVE_NP = True
except ImportError:
    HAVE_NP = False

try:
    from PIL import Image, features

    HAVE_PIL = True
    HAVE_JPEG = features.check_codec("jpg")
    HAVE_JPEG2K = features.check_codec("jpg_2000")
except ImportError:
    HAVE_PIL = False
    HAVE_JPEG = False
    HAVE_JPEG2K = False

from pydicom import config
from pydicom.encaps import generate_frames
from pydicom.misc import warn_and_log
from pydicom.pixels.utils import (
    pixel_dtype,
    get_j2k_parameters,
    get_nr_frames,
)
from pydicom.uid import (
    UID,
    JPEG2000,
    JPEG2000Lossless,
    JPEGBaseline8Bit,
    JPEGExtended12Bit,
)


logger = logging.getLogger("pydicom")


PillowJPEG2000TransferSyntaxes = [JPEG2000, JPEG2000Lossless]
PillowJPEGTransferSyntaxes = [JPEGBaseline8Bit, JPEGExtended12Bit]
PillowSupportedTransferSyntaxes = (
    PillowJPEGTransferSyntaxes + PillowJPEG2000TransferSyntaxes
)


HANDLER_NAME = "Pillow"
DEPENDENCIES = {
    "numpy": ("https://numpy.org/", "NumPy"),
    "PIL": ("https://python-pillow.org/", "Pillow"),
}


def is_available() -> bool:
    """Return ``True`` if the handler has its dependencies met."""
    return HAVE_NP and HAVE_PIL


def supports_transfer_syntax(transfer_syntax: UID) -> bool:
    """Return ``True`` if the handler supports the `transfer_syntax`.

    Parameters
    ----------
    transfer_syntax : uid.UID
        The Transfer Syntax UID of the *Pixel Data* that is to be used with
        the handler.
    """
    return transfer_syntax in PillowSupportedTransferSyntaxes


def needs_to_convert_to_RGB(ds: "Dataset") -> bool:
    """Return ``True`` if the *Pixel Data* should to be converted from YCbCr to
    RGB.

    This affects JPEG transfer syntaxes.
    """
    return False


def should_change_PhotometricInterpretation_to_RGB(ds: "Dataset") -> bool:
    """Return ``True`` if the *Photometric Interpretation* should be changed
    to RGB.

    This affects JPEG transfer syntaxes.
    """
    # return ds.SamplesPerPixel == 3
    return False


def _decompress_single_frame(
    data: bytes, transfer_syntax: str, photometric_interpretation: str
) -> "Image":
    """Decompresses a single frame of an encapsulated Pixel Data element.

    Parameters
    ----------
    data: bytes
        Compressed pixel data
    transfer_syntax: str
        Transfer Syntax UID
    photometric_interpretation: str
        Photometric Interpretation

    Returns
    -------
    PIL.Image
        Decompressed pixel data

    """
    fio = io.BytesIO(data)
    image = Image.open(fio)
    # This hack ensures that RGB color images, which were not
    # color transformed (i.e. not transformed into YCbCr color space)
    # upon JPEG compression are decompressed correctly.
    # Since Pillow assumes that images were transformed into YCbCr color
    # space prior to compression, setting the value of "mode" to YCbCr
    # signals Pillow to not apply any color transformation upon
    # decompression.
    if (
        transfer_syntax in PillowJPEGTransferSyntaxes
        and photometric_interpretation == "RGB"
        and "adobe_transform" not in image.info
    ):
        image.draft("YCbCr", image.size)
    return image


def get_pixeldata(ds: "Dataset") -> "numpy.ndarray":
    """Return a :class:`numpy.ndarray` of the *Pixel Data*.

    Parameters
    ----------
    ds : Dataset
        The :class:`Dataset` containing an Image Pixel module and the
        *Pixel Data* to be decompressed and returned.

    Returns
    -------
    numpy.ndarray
       The contents of (7FE0,0010) *Pixel Data* as a 1D array.

    Raises
    ------
    ImportError
        If Pillow is not available.
    NotImplementedError
        If the transfer syntax is not supported
    """
    transfer_syntax = ds.file_meta.TransferSyntaxUID

    if not HAVE_PIL:
        raise ImportError(
            f"The pillow package is required to use pixel_array for "
            f"this transfer syntax {transfer_syntax.name}, and pillow could "
            f"not be imported."
        )

    if not HAVE_JPEG and transfer_syntax in PillowJPEGTransferSyntaxes:
        raise NotImplementedError(
            f"The pixel data with transfer syntax {transfer_syntax.name}, "
            f"cannot be read because Pillow lacks the JPEG plugin"
        )

    if not HAVE_JPEG2K and transfer_syntax in PillowJPEG2000TransferSyntaxes:
        raise NotImplementedError(
            f"The pixel data with transfer syntax {transfer_syntax.name}, "
            f"cannot be read because Pillow lacks the JPEG 2000 plugin"
        )

    if transfer_syntax == JPEGExtended12Bit and ds.BitsAllocated != 8:
        raise NotImplementedError(
            f"{JPEGExtended12Bit} - {JPEGExtended12Bit.name} is only supported "
            "by Pillow if (0028,0100) Bits Allocated = 8"
        )

    photometric_interpretation = cast(str, ds.PhotometricInterpretation)
    rows = cast(int, ds.Rows)
    columns = cast(int, ds.Columns)
    bits_stored = cast(int, ds.BitsStored)
    bits_allocated = cast(int, ds.BitsAllocated)
    nr_frames = get_nr_frames(ds, warn=False)

    pixel_bytes = bytearray()
    j2k_precision, j2k_sign = None, None
    for frame in generate_frames(ds.PixelData, number_of_frames=nr_frames):
        im = _decompress_single_frame(
            frame, transfer_syntax, photometric_interpretation
        )
        if "YBR" in photometric_interpretation:
            im.draft("YCbCr", (rows, columns))
        pixel_bytes.extend(im.tobytes())

        if not j2k_precision:
            params = get_j2k_parameters(frame)
            j2k_precision = cast(int, params.setdefault("precision", bits_stored))
            j2k_sign = params.setdefault("is_signed", None)

    logger.debug(f"Successfully read {len(pixel_bytes)} pixel bytes")

    arr = numpy.frombuffer(pixel_bytes, pixel_dtype(ds))

    if transfer_syntax in PillowJPEG2000TransferSyntaxes:
        # Pillow converts N-bit data to 8- or 16-bit unsigned data,
        # See Pillow src/libImaging/Jpeg2KDecode.c::j2ku_gray_i
        shift = bits_allocated - bits_stored
        if j2k_precision and j2k_precision != bits_stored:
            warn_and_log(
                f"The (0028,0101) 'Bits Stored' value ({bits_stored}-bit) "
                f"doesn't match the JPEG 2000 data ({j2k_precision}-bit). "
                f"It's recommended that you change the 'Bits Stored' value"
            )

        if config.APPLY_J2K_CORRECTIONS and j2k_precision:
            # Corrections based on J2K data
            shift = bits_allocated - j2k_precision
            if not j2k_sign and j2k_sign != ds.PixelRepresentation:
                # Convert unsigned J2K data to 2's complement
                arr = numpy.right_shift(arr, shift)
            else:
                if ds.PixelRepresentation == 1:
                    # Pillow converts signed data to unsigned
                    #   so we need to undo this conversion
                    arr -= numpy.uint32(2 ** (bits_allocated - 1))

                if shift:
                    arr = numpy.right_shift(arr, shift)
        else:
            # Corrections based on dataset elements
            if ds.PixelRepresentation == 1:
                arr -= numpy.uint32(2 ** (bits_allocated - 1))

            if shift:
                arr = numpy.right_shift(arr, shift)

    if should_change_PhotometricInterpretation_to_RGB(ds):
        ds.PhotometricInterpretation = "RGB"

    return cast("numpy.ndarray", arr)



# Copyright 2008-2020 pydicom authors. See LICENSE file for details.
"""Define the Sequence class, which contains a sequence DataElement's items.

Sequence is a list of pydicom Dataset objects.
"""
from typing import cast, Any, TypeVar
from collections.abc import Iterable

from pydicom.dataset import Dataset
from pydicom.multival import ConstrainedList


# Python 3.11 adds typing.Self, until then...
Self = TypeVar("Self", bound="Sequence")


class Sequence(ConstrainedList[Dataset]):
    """Class to hold multiple :class:`~pydicom.dataset.Dataset` in a :class:`list`."""

    def __init__(self, iterable: Iterable[Dataset] | None = None) -> None:
        """Initialize a list of :class:`~pydicom.dataset.Dataset`.

        Parameters
        ----------
        iterable : Iterable[Dataset] | None
            An iterable object (e.g. :class:`list`, :class:`tuple`) containing
            :class:`~pydicom.dataset.Dataset`. If not used then an empty
            :class:`Sequence` is generated.
        """
        # We add this extra check to throw a relevant error. Without it, the
        # error will be simply that a Sequence must contain Datasets (since a
        # Dataset IS iterable). This error, however, doesn't inform the user
        # that the actual issue is that their Dataset needs to be INSIDE an
        # iterable object
        if isinstance(iterable, Dataset):
            raise TypeError("The Sequence constructor requires an iterable")

        # If True, SQ element uses an undefined length of 0xFFFFFFFF
        self.is_undefined_length: bool

        super().__init__(iterable)

    def extend(self, val: Iterable[Dataset]) -> None:
        """Extend the :class:`~pydicom.sequence.Sequence` using an iterable
        of :class:`~pydicom.dataset.Dataset` instances.
        """
        if isinstance(val, Dataset):
            raise TypeError("An iterable of 'Dataset' is required")

        super().extend(val)

    def __iadd__(self: Self, other: Iterable[Dataset]) -> Self:
        """Implement Sequence() += [Dataset()]."""
        if isinstance(other, Dataset):
            raise TypeError("An iterable of 'Dataset' is required")

        return super().__iadd__(other)

    def __setitem__(self, index: slice | int, val: Iterable[Dataset] | Dataset) -> None:
        """Add item(s) to the Sequence at `index`."""
        if isinstance(index, slice):
            if isinstance(val, Dataset):
                raise TypeError("Can only assign an iterable of 'Dataset'")

            super().__setitem__(index, val)
        else:
            super().__setitem__(index, cast(Dataset, val))

    def __str__(self) -> str:
        """String description of the Sequence."""
        return f"[{''.join([str(x) for x in self])}]"

    def __repr__(self) -> str:
        """String representation of the Sequence."""
        return f"<{self.__class__.__name__}, length {len(self)}>"

    @staticmethod
    def _validate(item: Any) -> Dataset:
        """Check that `item` is a :class:`~pydicom.dataset.Dataset` instance."""
        if isinstance(item, Dataset):
            return item

        raise TypeError("Sequence contents must be 'Dataset' instances.")



# Copyright 2008-2018 pydicom authors. See LICENSE file for details.
"""
Use the `jpeg_ls (pyjpegls) <https://github.com/pydicom/pyjpegls>`_ Python
package to decode *Pixel Data*.
"""
from typing import TYPE_CHECKING, cast

try:
    import numpy

    HAVE_NP = True
except ImportError:
    HAVE_NP = False

try:
    import jpeg_ls

    HAVE_JPEGLS = True
except ImportError:
    HAVE_JPEGLS = False

from pydicom.encaps import generate_frames
from pydicom.pixels.utils import pixel_dtype, get_nr_frames
import pydicom.uid

if TYPE_CHECKING:  # pragma: no cover
    from pydicom.dataset import Dataset


HANDLER_NAME = "JPEG-LS"

DEPENDENCIES = {
    "numpy": ("https://numpy.org/", "NumPy"),
    "jpeg_ls": ("https://github.com/pydicom/pyjpegls", "pyjpegls"),
}

SUPPORTED_TRANSFER_SYNTAXES = [
    pydicom.uid.JPEGLSLossless,
    pydicom.uid.JPEGLSNearLossless,
]


def is_available() -> bool:
    """Return ``True`` if the handler has its dependencies met."""
    return HAVE_NP and HAVE_JPEGLS


def needs_to_convert_to_RGB(ds: "Dataset") -> bool:
    """Return ``True`` if the *Pixel Data* should to be converted from YCbCr to
    RGB.

    This affects JPEG transfer syntaxes.
    """
    return False


def should_change_PhotometricInterpretation_to_RGB(ds: "Dataset") -> bool:
    """Return ``True`` if the *Photometric Interpretation* should be changed
    to RGB.

    This affects JPEG transfer syntaxes.
    """
    return False


def supports_transfer_syntax(transfer_syntax: pydicom.uid.UID) -> bool:
    """Return ``True`` if the handler supports the `transfer_syntax`.

    Parameters
    ----------
    transfer_syntax : uid.UID
        The Transfer Syntax UID of the *Pixel Data* that is to be used with
        the handler.
    """
    return transfer_syntax in SUPPORTED_TRANSFER_SYNTAXES


def get_pixeldata(ds: "Dataset") -> "numpy.ndarray":
    """Return the *Pixel Data* as a :class:`numpy.ndarray`.

    Returns
    -------
    numpy.ndarray
        A correctly sized (but not shaped) numpy array of the *Pixel Data*.

    Raises
    ------
    ImportError
        If the required packages are not available.
    NotImplementedError
        If the transfer syntax is not supported.
    TypeError
        If the pixel data type is unsupported.
    """
    tsyntax = ds.file_meta.TransferSyntaxUID
    if tsyntax not in SUPPORTED_TRANSFER_SYNTAXES:
        raise NotImplementedError(
            f"The jpeg_ls does not support this transfer syntax {tsyntax.name}"
        )

    if not HAVE_JPEGLS:
        raise ImportError(
            "The jpeg_ls package is required to use pixel_array for this "
            f"transfer syntax {tsyntax.name}, and jpeg_ls could not be "
            "imported"
        )

    pixel_bytes = bytearray()

    nr_frames = get_nr_frames(ds, warn=False)
    for frame in generate_frames(ds.PixelData, number_of_frames=nr_frames):
        im = jpeg_ls.decode(numpy.frombuffer(frame, dtype="u1"))
        pixel_bytes.extend(im.tobytes())

    arr = numpy.frombuffer(pixel_bytes, pixel_dtype(ds))

    if should_change_PhotometricInterpretation_to_RGB(ds):
        ds.PhotometricInterpretation = "RGB"

    return cast("numpy.ndarray", arr)



# Copyright 2008-2020 pydicom authors. See LICENSE file for details.
"""Functions for working with encapsulated (compressed) pixel data."""

from collections.abc import Iterator
from io import BytesIO, BufferedIOBase
import os
from struct import pack, unpack
from typing import Any

from pydicom import config
from pydicom.misc import warn_and_log
from pydicom.filebase import DicomBytesIO, DicomIO, ReadableBuffer
from pydicom.fileutil import buffer_length, reset_buffer_position
from pydicom.tag import Tag, ItemTag, SequenceDelimiterTag


# Functions for parsing encapsulated data
def parse_basic_offsets(
    buffer: bytes | bytearray | ReadableBuffer, *, endianness: str = "<"
) -> list[int]:
    """Return the encapsulated pixel data's basic offset table frame offsets.

    .. versionadded:: 3.0

    Parameters
    ----------
    buffer : bytes | bytearray | readable buffer
        A buffer containing the encapsulated frame data, positioned at the
        beginning of the Basic Offset Table. May be :class:`bytes`,
        :class:`bytearray` or an object with ``read()``, ``tell()`` and
        ``seek()`` methods. If the latter then after reading it will be
        positioned at the start of the item tag of the first fragment after the
        Basic Offset Table.
    endianness : str, optional
        If ``"<"`` (default) then the encapsulated data uses little endian
        encoding, otherwise if ``">"`` it uses big endian encoding.

    Returns
    -------
    list[int]
        A list of the offset positions to the first item tag of each frame, as
        measured from the end of the basic offset table.

    References
    ----------
    :dcm:`DICOM Standard, Part 5, Annex A.4<part05/sect_A.4.html#table_A.4-1>`
    """
    if isinstance(buffer, bytes | bytearray):
        buffer = BytesIO(buffer)

    group, elem = unpack(f"{endianness}HH", buffer.read(4))
    if group << 16 | elem != 0xFFFEE000:
        raise ValueError(
            f"Found unexpected tag {Tag(group, elem)} instead of (FFFE,E000) "
            "when parsing the Basic Offset Table item"
        )

    length = unpack(f"{endianness}L", buffer.read(4))[0]
    if length % 4:
        raise ValueError(
            "The length of the Basic Offset Table item is not a multiple of 4"
        )

    if length == 0:
        return []

    return list(unpack(f"{endianness}{length // 4}L", buffer.read(length)))


def parse_fragments(
    buffer: bytes | bytearray | ReadableBuffer, *, endianness: str = "<"
) -> tuple[int, list[int]]:
    """Return the number of fragments and their positions in `buffer`.

    .. versionadded:: 3.0

    Parameters
    ----------
    buffer : bytes | bytearray | readable buffer
        A buffer containing the encapsulated frame data, starting at the first
        byte of item tag for a fragment, such as after the end of the Basic
        Basic Offset Table. May be :class:`bytes`, :class:`bytearray` or an
        object with ``read()``, ``tell()`` and ``seek()`` methods. If the latter
        then the offset will be reset to the starting position afterwards.
    endianness : str, optional
        If ``"<"`` (default) then the encapsulated data uses little endian
        encoding, otherwise if ``">"`` it uses big endian encoding.

    Returns
    -------
    tuple[int, list[int]]
        The number of fragments and the absolute offset position of the first
        byte of the item tag for each fragment in `buffer`.
    """
    if isinstance(buffer, bytes | bytearray):
        buffer = BytesIO(buffer)

    start_offset = buffer.tell()

    nr_fragments = 0
    fragment_offsets = []
    while True:
        try:
            group, elem = unpack(f"{endianness}HH", buffer.read(4))
        except Exception:
            break

        tag = group << 16 | elem
        if tag == 0xFFFEE000:
            if len(raw_length := buffer.read(4)) != 4:
                raise ValueError(
                    "Unable to determine the length of the item at offset "
                    f"{buffer.tell() - len(raw_length) - 4} as the end of "
                    "the data has been reached - the encapsulated pixel data "
                    "may be invalid"
                )
            length = unpack(f"{endianness}L", raw_length)[0]
            if length == 0xFFFFFFFF:
                raise ValueError(
                    f"Undefined item length at offset {buffer.tell() - 4} when "
                    "parsing the encapsulated pixel data fragments"
                )
            nr_fragments += 1
            fragment_offsets.append(buffer.tell() - 8)
            buffer.seek(length, 1)
        elif tag == 0xFFFEE0DD:
            break
        else:
            raise ValueError(
                f"Unexpected tag '{Tag(tag)}' at offset {buffer.tell() - 4} when "
                "parsing the encapsulated pixel data fragment items"
            )

    buffer.seek(start_offset, 0)

    return nr_fragments, fragment_offsets


def generate_fragments(
    buffer: bytes | bytearray | ReadableBuffer, *, endianness: str = "<"
) -> Iterator[bytes]:
    """Yield frame fragments from the encapsulated pixel data in `buffer`.

    .. versionadded:: 3.0

    Parameters
    ----------
    buffer : bytes | bytearray | readable buffer
        A buffer containing the encapsulated frame data, starting at the first
        byte of item tag for a fragment, usually this will be after the end
        of the Basic Offset Table. May be :class:`bytes`, :class:`bytearray` or
        an object with ``read()``, ``tell()`` and ``seek()`` methods. If the
        latter than the final offset position depends on how many fragments have
        been yielded.
    endianness : str, optional
        If ``"<"`` (default) then the encapsulated data uses little endian
        encoding, otherwise if ``">"`` it uses big endian encoding.

    Yields
    ------
    bytes
        A pixel data fragment.
    """
    if isinstance(buffer, bytes | bytearray):
        buffer = BytesIO(buffer)

    while True:
        try:
            group, elem = unpack(f"{endianness}HH", buffer.read(4))
        except Exception:
            break

        tag = group << 16 | elem
        if tag == 0xFFFEE000:
            if len(raw_length := buffer.read(4)) != 4:
                raise ValueError(
                    "Unable to determine the length of the item at offset "
                    f"{buffer.tell() - len(raw_length) - 4} as the end of "
                    "the data has been reached - the encapsulated pixel data "
                    "may be invalid"
                )
            length = unpack(f"{endianness}L", raw_length)[0]
            if length == 0xFFFFFFFF:
                raise ValueError(
                    f"Undefined item length at offset {buffer.tell() - 4} when "
                    "parsing the encapsulated pixel data fragments"
                )

            yield buffer.read(length)
        elif tag == 0xFFFEE0DD:
            break
        else:
            raise ValueError(
                f"Unexpected tag '{Tag(tag)}' at offset {buffer.tell() - 4} when "
                "parsing the encapsulated pixel data fragment items"
            )


def generate_fragmented_frames(
    buffer: bytes | bytearray | ReadableBuffer,
    *,
    number_of_frames: int | None = None,
    extended_offsets: tuple[list[int], list[int]] | tuple[bytes, bytes] | None = None,
    endianness: str = "<",
) -> Iterator[tuple[bytes, ...]]:
    """Yield fragmented pixel data frames from `buffer`.

    .. versionadded:: 3.0

    .. note::

        When the Basic Offset Table is empty and the Extended Offset Table
        isn't supplied then more fragmented frames may be yielded than given
        by `number_of_frames` provided there are sufficient excess fragments
        available.

    Parameters
    ----------
    buffer : bytes | bytearray | readable buffer
        A buffer containing the encapsulated frame data, positioned at the first
        byte of the basic offset table. May be :class:`bytes`,
        :class:`bytearray` or an object with ``read()``, ``tell()`` and
        ``seek()`` methods. If the latter then the final position depends on
        how many fragmented frames have been yielded.
    number_of_frames : int, optional
        Required when the Basic Offset Table is empty and the Extended Offset Table
        has not been supplied. This should be the value of (0028,0008) *Number of
        Frames* or the expected number of frames in the encapsulated data.
    extended_offsets : tuple[list[int], list[int]] or tuple[bytes, bytes], optional
        The (offsets, lengths) of the Extended Offset Table as taken from
        (7FE0,0001) *Extended Offset Table* and (7FE0,0002) *Extended Offset
        Table Lengths* as either the raw encoded values or a list of their
        decoded equivalents.
    endianness : str, optional
        If ``"<"`` (default) then the encapsulated data uses little endian
        encoding, otherwise if ``">"`` it uses big endian encoding.

    Yields
    -------
    tuple[bytes, ...]
        An encapsulated pixel data frame, with the contents of the tuple the
        frame's fragmented encoded data.
    """
    if isinstance(buffer, bytes | bytearray):
        buffer = BytesIO(buffer)

    basic_offsets = parse_basic_offsets(buffer, endianness=endianness)
    # `buffer` is positioned at the end of the basic offsets table

    # Prefer the extended offset table (if available)
    if extended_offsets:
        # PS3.3, C.7.6.3.1.8
        # Byte offsets to the first byte of the item tag of the first fragment
        #   of every frame, as measured from the first byte of the item tag
        #   following the Basic Offset Table, which *should* be empty
        # Only 1 fragment per frame is allowed (Table C.7-11a)
        if isinstance(extended_offsets[0], bytes):
            nr_offsets = len(extended_offsets[0]) // 8
            offsets = list(unpack(f"{endianness}{nr_offsets}Q", extended_offsets[0]))
        else:
            offsets = extended_offsets[0]

        if isinstance(extended_offsets[1], bytes):
            nr_offsets = len(extended_offsets[1]) // 8
            lengths = list(unpack(f"{endianness}{nr_offsets}Q", extended_offsets[1]))
        else:
            lengths = extended_offsets[1]

        fragments_start = buffer.tell()
        for offset, length in zip(offsets, lengths):
            # 8 for the item tag and item length, which we don't need
            buffer.seek(fragments_start + offset + 8, 0)
            yield (buffer.read(length),)

        return

    # Fall back to the basic offset table (if available)
    if basic_offsets:
        frame = []
        current_index = 0
        current_offset = 0
        final_index = len(basic_offsets) - 1
        for fragment in generate_fragments(buffer, endianness=endianness):
            if current_index == final_index:
                # Nth frame, keep adding fragments until we have no more
                frame.append(fragment)
                continue

            if current_offset < basic_offsets[current_index + 1]:
                # N - 1th frame, keep adding fragments until the we go
                #   past the next frame offset
                frame.append(fragment)
            else:
                # Gone past the next offset, yield and restart
                yield tuple(frame)
                current_index += 1
                frame = [fragment]

            # + 8 bytes for item tag and item length
            current_offset += len(fragment) + 8

        # Yield the Nth frame
        yield tuple(frame)
        return

    # No basic or extended offset table
    # Determine the number of fragments in the buffer
    nr_fragments, _ = parse_fragments(buffer, endianness=endianness)
    # `buffer` is positioned at the end of the basic offsets table
    fragments = generate_fragments(buffer, endianness=endianness)

    # Single fragment must be 1 frame
    if nr_fragments == 1:
        yield (next(fragments),)
        return

    # From this point on we require the number of frames as there are
    #   multiple fragments and may be one or more frames
    if not number_of_frames:
        raise ValueError(
            "Unable to determine the frame boundaries for the encapsulated "
            "pixel data as there is no Basic or Extended Offset Table and "
            "the number of frames has not been supplied"
        )

    # 1 fragment per frame, for N frames
    if nr_fragments == number_of_frames:
        # Covers RLE and others if 1:1 ratio
        yield from ((fragment,) for fragment in fragments)
        return

    # Multiple fragments for 1 frame
    if number_of_frames == 1:
        yield tuple(fragment for fragment in fragments)
        return

    # More fragments then frames
    if nr_fragments > number_of_frames:
        # Search for JPEG/JPEG-LS/JPEG2K EOI/EOC marker which should be the
        #   last two bytes of a frame
        # It's possible to yield more frames than `number_of_frames` as
        #   long as there are excess fragments with JPEG EOI/EOC markers
        # It's also possible that we yielded too early because the marker bytes
        #   were actually part of the compressed JPEG codestream
        eoi_marker = b"\xff\xd9"
        frame = []
        frame_nr = 0
        for fragment in fragments:
            frame.append(fragment)
            if eoi_marker in fragment[-10:]:
                yield tuple(frame)
                frame_nr += 1
                frame = []

        # There was a final set of fragments with no EOI/EOC marker, data is
        #   probably corrupted, but yield it and warn/log anyway
        if frame:
            if frame_nr >= number_of_frames:
                msg = (
                    "The end of the encapsulated pixel data has been reached but "
                    "no JPEG EOI/EOC marker was found, the final frame may be "
                    "be invalid"
                )
            else:
                msg = (
                    "The end of the encapsulated pixel data has been reached but "
                    "fewer frames than expected have been found. Please confirm "
                    "that the generated frame data is correct"
                )

            warn_and_log(msg)
            yield tuple(frame)

        elif frame_nr < number_of_frames:
            warn_and_log(
                "The end of the encapsulated pixel data has been reached but "
                "fewer frames than expected have been found"
            )

        return

    # nr_fragments < number_of_frames
    raise ValueError(
        "Unable to generate frames from the encapsulated pixel data as there "
        "are fewer fragments than frames; the dataset may be corrupt or the "
        "number of frames may be incorrect"
    )


def generate_frames(
    buffer: bytes | ReadableBuffer,
    *,
    number_of_frames: int | None = None,
    extended_offsets: tuple[list[int], list[int]] | tuple[bytes, bytes] | None = None,
    endianness: str = "<",
) -> Iterator[bytes]:
    """Yield complete pixel data frames from `buffer`.

    .. versionadded:: 3.0

    .. note::

        When the Basic Offset Table is empty and the Extended Offset Table
        isn't supplied then more frames may be yielded than given by
        `number_of_frames` provided there are sufficient excess fragments
        available.

    Parameters
    ----------
    buffer : bytes | readable buffer
        A buffer containing the encapsulated frame data, starting at the first
        byte of the basic offset table. May be :class:`bytes`,
        :class:`bytearray` or an object with ``read()``, ``tell()`` and
        ``seek()`` methods. If the latter then the final offset position depends
        on the number of yielded frames.
    number_of_frames : int, optional
        Required when the Basic Offset Table is empty and the Extended Offset
        Table has not been supplied. This should be the value of (0028,0008) *Number
        of Frames* or the expected number of frames in the encapsulated data.
    extended_offsets : tuple[list[int], list[int]] or tuple[bytes, bytes], optional
        The (offsets, lengths) of the Extended Offset Table as taken from
        (7FE0,0001) *Extended Offset Table* and (7FE0,0002) *Extended Offset
        Table Lengths* as either the raw encoded values or a list of their
        decoded equivalents.
    endianness : str, optional
        If ``"<"`` (default) then the encapsulated data uses little endian
        encoding, otherwise if ``">"`` it uses big endian encoding.

    Yields
    ------
    bytes
        The encoded pixel data, one frame at a time.

    References
    ----------
    DICOM Standard Part 5, :dcm:`Annex A <part05/chapter_A.html>`
    """
    fragmented_frames = generate_fragmented_frames(
        buffer,
        number_of_frames=number_of_frames,
        extended_offsets=extended_offsets,
        endianness=endianness,
    )
    for fragments in fragmented_frames:
        yield b"".join(fragments)


def get_frame(
    buffer: bytes | bytearray | ReadableBuffer,
    index: int,
    *,
    extended_offsets: tuple[list[int], list[int]] | tuple[bytes, bytes] | None = None,
    number_of_frames: int | None = None,
    endianness: str = "<",
) -> bytes:
    """Return the specified frame at `index`.

    .. versionadded:: 3.0

    .. note::

        When the Basic Offset Table is empty and the Extended Offset Table
        isn't supplied then it's possible to return a frame at a higher `index`
        than expected from the supplied `number_of_frames` value provided there
        are sufficient excess fragments available.

    Parameters
    ----------
    buffer : bytes | bytearray | readable buffer
        A buffer containing the encapsulated frame data, positioned at the first
        byte of the basic offset table. May be :class:`bytes`,
        :class:`bytearray` or an object with ``read()``, ``tell()`` and
        ``seek()`` methods. If the latter then the buffer will be reset to the
        starting position if the frame was
        returned successfully.
    index : int
        The index of the frame to be returned, starting at ``0`` for the first
        frame.
    number_of_frames : int, optional
        Required for multi-frame data when the Basic Offset Table is empty,
        the Extended Offset Table has not been supplied and there are
        multiple frames. This should be the value of (0028,0008) *Number of
        Frames* or the expected number of frames in the encapsulated data.
    extended_offsets : tuple[list[int], list[int]] or tuple[bytes, bytes], optional
        The (offsets, lengths) of the Extended Offset Table as taken from
        (7FE0,0001) *Extended Offset Table* and (7FE0,0002) *Extended Offset
        Table Lengths* as either the raw encoded values or a list of their
        decoded equivalents.
    endianness : str, optional
        If ``"<"`` (default) then the encapsulated data uses little endian
        encoding, otherwise if ``">"`` it uses big endian encoding.

    Returns
    -------
    bytes
        A single frame of encoded pixel data.


    References
    ----------
    DICOM Standard Part 5, :dcm:`Annex A <part05/chapter_A.html>`
    """
    if isinstance(buffer, bytes | bytearray):
        buffer = BytesIO(buffer)

    # `buffer` is positioned at the start of the basic offsets table
    starting_position = buffer.tell()

    basic_offsets = parse_basic_offsets(buffer, endianness=endianness)
    # `buffer` is positioned at the end of the basic offsets table

    # Prefer the extended offset table (if available)
    if extended_offsets:
        if isinstance(extended_offsets[0], bytes):
            nr_offsets = len(extended_offsets[0]) // 8
            offsets = list(unpack(f"{endianness}{nr_offsets}Q", extended_offsets[0]))
        else:
            offsets = extended_offsets[0]

        if isinstance(extended_offsets[1], bytes):
            nr_offsets = len(extended_offsets[1]) // 8
            lengths = list(unpack(f"{endianness}{nr_offsets}Q", extended_offsets[1]))
        else:
            lengths = extended_offsets[1]

        if index >= len(offsets):
            raise ValueError(
                "There aren't enough offsets in the Extended Offset Table for "
                f"{index + 1} frames"
            )

        # We have the length so skip past the item tag and item length
        buffer.seek(offsets[index] + 8, 1)
        frame = buffer.read(lengths[index])
        buffer.seek(starting_position)
        return frame

    # Fall back to the basic offset table (if available)
    if basic_offsets:
        if index >= len(basic_offsets):
            raise ValueError(
                "There aren't enough offsets in the Basic Offset Table for "
                f"{index + 1} frames"
            )

        # There may be multiple fragments per frame
        if index < len(basic_offsets) - 1:
            # N - 1th frames
            length = basic_offsets[index + 1] - basic_offsets[index]
            buffer.seek(basic_offsets[index], 1)
            fragments = generate_fragments(buffer.read(length), endianness=endianness)
        else:
            # Final frame
            buffer.seek(basic_offsets[-1], 1)
            fragments = generate_fragments(buffer, endianness=endianness)

        frame = b"".join(fragment for fragment in fragments)
        buffer.seek(starting_position)
        return frame

    # No basic or extended offset table
    # Determine the number of fragments in `buffer` an their offsets
    nr_fragments, fragment_offsets = parse_fragments(buffer, endianness=endianness)
    # `buffer` is positioned at the end of the basic offsets table

    # Single fragment must be 1 frame
    if nr_fragments == 1:
        if index == 0:
            frame = next(generate_fragments(buffer, endianness=endianness))
            buffer.seek(starting_position, 0)
            return frame

        raise ValueError(
            "Found 1 frame fragment in the encapsulated pixel data, 'index' must be 0"
        )

    # From this point on we require the number of frames as there are
    #   multiple fragments and may be one or more frames
    if not number_of_frames:
        raise ValueError(
            "Unable to determine the frame boundaries for the encapsulated "
            "pixel data as there is no basic or extended offset table data and "
            "the number of frames has not been supplied"
        )

    # 1 fragment per frame, for N frames
    if nr_fragments == number_of_frames:
        if index > nr_fragments - 1:
            raise ValueError(
                f"Found {nr_fragments} frame fragments in the encapsulated "
                f"pixel data, an 'index' of {index} is invalid"
            )

        # Covers RLE and others if 1:1 ratio
        # `fragment_offsets` is the absolute positions of each item tag
        buffer.seek(fragment_offsets[index], 0)
        frame = next(generate_fragments(buffer, endianness=endianness))
        buffer.seek(starting_position, 0)
        return frame

    fragments = generate_fragments(buffer, endianness=endianness)

    # Multiple fragments for 1 frame
    if number_of_frames == 1:
        if index == 0:
            frame = b"".join(fragment for fragment in fragments)
            buffer.seek(starting_position, 0)
            return frame

        raise ValueError("The 'index' must be 0 if the number of frames is 1")

    # Search for JPEG/JPEG-LS/JPEG2K EOI/EOC marker which should be the
    #   last two bytes of a frame
    eoi_marker = b"\xFF\xD9"
    frame_fragments = []
    frame_nr = 0
    for fragment in fragments:
        frame_fragments.append(fragment)
        if eoi_marker in fragment[-10:]:
            if frame_nr == index:
                frame = b"".join(frame_fragments)
                buffer.seek(starting_position, 0)
                return frame

            frame_nr += 1
            frame_fragments = []

    if frame_fragments and index == frame_nr:
        warn_and_log(
            "The end of the encapsulated pixel data has been reached but no "
            "JPEG EOI/EOC marker was found, the returned frame data may be "
            "invalid"
        )
        frame = b"".join(frame_fragments)
        buffer.seek(starting_position, 0)
        return frame

    raise ValueError(f"There is insufficient pixel data to contain {index + 1} frames")


# Functions and classes for encapsulating data
class _BufferedItem:
    """Convenience class for a buffered encapsulation item.

    Attributes
    ----------
    buffer : io.BufferedIOBase
        The buffer containing data to be encapsulated.
    length : int
        The total length of the encapsulated item, including the item tag and
        value and any trailing padding required to bring the item data up to an
        even length.
    """

    def __init__(self, buffer: BufferedIOBase) -> None:
        """Create a new ``_BufferedItem`` instance.

        Parameters
        ----------
        buffer : io.BufferedIOBase
            The buffer containing data to be encapsulated, may be empty.
        """
        self.buffer = buffer
        # The non-padded length of the data in the buffer
        self._blen = buffer_length(buffer)

        if self._blen > 2**32 - 2:
            raise ValueError(
                "Buffers containing more than 4294967294 bytes are not supported"
            )

        # 8 bytes for the item tag and length
        self.length = 8 + self._blen + self._blen % 2
        # Whether or not the buffer needs trailing padding
        self._padding = bool(self._blen % 2)
        # The item tag and length
        self._item = b"".join(
            (
                b"\xFE\xFF\x00\xE0",
                (self.length - 8).to_bytes(length=4, byteorder="little"),
            )
        )

    def read(self, start: int, size: int) -> bytes:
        """Return data from the encapsulated frame.

        Parameters
        ----------
        start : int
            The initial position in the encapsulated item where data should be read
            from, must be greater than or equal to 0, where ``0`` is the first byte
            of the item tag.
        size : int
            The number of bytes to read from the buffer.

        Returns
        -------
        bytes
            The data read from the buffer.
        """
        if not 0 <= start < self.length:
            raise ValueError(
                f"Invalid 'start' value '{start}', must be in the closed interval "
                f"[0, {self.length - 1}]"
            )

        nr_read = 0
        out = bytearray()
        while length := (size - nr_read):
            offset = start + nr_read
            if offset < 8:
                # `offset` in item tag/length
                _read = self._item[offset : offset + length]
            elif 0 <= (offset - 8) < self._blen:
                # `offset` in item value
                with reset_buffer_position(self.buffer):
                    self.buffer.seek(offset - 8)
                    _read = self.buffer.read(length)

            elif self._padding and offset == self.length - 1:
                # `offset` in the item value padding
                _read = b"\x00"
            else:
                # `offset` past the end of the item value
                _read = b""

            if not _read:
                break

            nr_read += len(_read)
            out.extend(_read)

        return bytes(out)


class EncapsulatedBuffer(BufferedIOBase):
    """Convenience class for managing the encapsulation of one or more buffers
    containing compressed *Pixel Data*.

    .. versionadded:: 3.0
    """

    def __init__(self, buffers: list[BufferedIOBase], use_bot: bool = False) -> None:
        """Create a new class instance.

        Parameters
        ----------
        buffers : list[io.BufferedIOBase]
            The buffered pixel data frames to be encapsulated on writing the dataset.
            Only a single frame of pixel data can be in each buffer and the buffers
            must inherit from :class:`io.BufferedIODBase` and be readable and seekable.
        use_bot : bool, optional
            If ``True`` then the Basic Offset Table will include the offsets for
            each encapsulated items, otherwise no offsets will be included (default).
        """
        if not isinstance(buffers, list):
            raise TypeError(
                "'buffers' must be a list of objects that inherit from "
                "'io.BufferedIOBase'"
            )

        # The items to be encapsulated
        self._items = [_BufferedItem(b) for b in buffers]

        # The current position of the encapsulated buffer
        self._offset = 0

        # Use a non-empty Basic Offset Table
        self._use_bot = use_bot

        # The basic offset table
        bot = self.basic_offset_table

        # Offsets for the buffered items
        # Start of the item tag for the Basic Offset Table
        self._item_offsets = [0]
        # Start of the item tag for each frame
        self._item_offsets.extend([offset + len(bot) for offset in self.offsets])
        # End of the encapsulation
        self._item_offsets.append(self.encapsulated_length)

        # A dict containing the items to read encoded data from
        #   0: the buffered Basic Offset Table value
        #   1 to len(buffers): the buffered frames
        self._buffers = {idx + 1: item for idx, item in enumerate(self._items)}
        self._buffers[0] = _BufferedItem(BytesIO(bot[8:]))

    @property
    def basic_offset_table(self) -> bytes:
        """Return an encoded Basic Offset Table."""
        if not self._use_bot:
            return b"\xFE\xFF\x00\xE0\x00\x00\x00\x00"

        # The item tag for the offset table
        bot = [b"\xFE\xFF\x00\xE0"]
        # Add the item length
        bot.append(pack("<I", 4 * len(self.offsets)))
        # Add the item value
        bot.append(pack(f"<{len(self.offsets)}I", *self.offsets))

        return b"".join(bot)

    @property
    def closed(self) -> bool:
        """Return ``True`` if any of the encapsulated buffers are closed."""
        return any(item.buffer.closed for item in self._items)

    @property
    def extended_lengths(self) -> bytes:
        """Return an encoded *Extended Offset Table Lengths* value from `lengths`

        Returns
        -------
        bytes
            The encoded lengths of the frame.
        """
        # Exclude the item tag and item length
        return pack(f"<{len(self.lengths)}Q", *(length - 8 for length in self.lengths))

    @property
    def extended_offsets(self) -> bytes:
        """Return an encoded *Extended Offset Table* value from `offsets`

        Returns
        -------
        bytes
            The encoded offsets to the first byte of the item tag of the first fragment
            for every frame, as measured from the first byte of the first item tag
            following the empty Basic Offset Table Item.
        """
        return pack(f"<{len(self.offsets)}Q", *self.offsets)

    @property
    def encapsulated_length(self) -> int:
        """Return the total length of the encapulated *Pixel Data* value."""
        return len(self.basic_offset_table) + sum(self.lengths)

    @property
    def lengths(self) -> list[int]:
        """Return the encapsulated item lengths."""
        return [item.length for item in self._items]

    @property
    def offsets(self) -> list[int]:
        """Return the encapsulated item offsets, starting at 0 for the first item."""
        return [sum(self.lengths[0:idx]) for idx, _ in enumerate(self.lengths)]

    def read(self, size: int | None = 8192, /) -> bytes:
        """Read up to `size` bytes of data from the encapsulated buffers.

        Parameters
        ----------
        size : int, optional
            The amount of data to be read, if ``None`` then all data will be returned.

        Returns
        -------
        bytes
            The data read from the encapsulated buffers.
        """
        if self._offset >= self.encapsulated_length:
            return b""

        size = self.encapsulated_length if size is None else size

        nr_read = 0
        out = bytearray()
        while length := (size - nr_read):
            iterator = enumerate(zip(self._item_offsets, self._item_offsets[1:]))
            for idx, (start, end) in iterator:
                if start <= self._offset < end:
                    _read = self._buffers[idx].read(self._offset - start, length)
                    break

            if not _read:
                break

            self._offset += len(_read)
            nr_read += len(_read)
            out.extend(_read)

            if self._offset >= self.encapsulated_length:
                break

        return bytes(out)

    def readable(self) -> bool:
        """Return ``True`` if all the encapsulated buffers are readable."""
        return all(item.buffer.readable() for item in self._items)

    def seek(self, offset: int, whence: int = os.SEEK_SET, /) -> int:
        """Change the encapsulated buffers position to the given byte `offset`,
        relative to the position indicated by `whence` and return the new absolute
        position.
        """
        if whence not in (os.SEEK_SET, os.SEEK_CUR, os.SEEK_END):
            raise ValueError("Invalid 'whence' value, should be 0, 1 or 2")

        # Behavior emulates io.BytesIO
        if whence == os.SEEK_SET:
            # relative to beginning of buffer
            if offset < 0:
                raise ValueError(f"Negative seek 'offset' value {offset}")

            new_offset = offset
        elif whence == os.SEEK_CUR:
            # relative to current buffer position
            new_offset = self._offset + offset
            new_offset = 0 if new_offset < 0 else new_offset
        elif whence == os.SEEK_END:
            # relative to end of the buffer
            new_offset = self.encapsulated_length + offset
            new_offset = 0 if new_offset < 0 else new_offset

        self._offset = new_offset

        return self._offset

    def seekable(self) -> bool:
        """Return ``True`` if all the encapsulated buffers are seekable."""
        return all(item.buffer.seekable() for item in self._items)

    def tell(self) -> int:
        """Return the current stream position of the encapsulated buffers"""
        return self._offset


def fragment_frame(frame: bytes, nr_fragments: int = 1) -> Iterator[bytes]:
    """Yield one or more fragments from `frame`.

    .. versionadded:: 1.2

    Parameters
    ----------
    frame : bytes
        The data to fragment.
    nr_fragments : int, optional
        The number of fragments (default ``1``).

    Yields
    ------
    bytes
        The fragmented data, with all fragments as an even number of bytes
        greater than or equal to two.

    Notes
    -----

    * All items containing an encoded fragment shall be made of an even number
      of bytes greater than or equal to two.
    * The last fragment of a frame may be padded, if necessary to meet the
      sequence item format requirements of the DICOM Standard.
    * Any necessary padding may be appended after the end of image marker.
    * Encapsulated Pixel Data has the Value Representation OB.
    * Values with a VR of OB shall be padded with a single trailing NULL byte
      value (``0x00``) to achieve even length.

    References
    ----------
    DICOM Standard, Part 5, :dcm:`Section 6.2 <part05/sect_6.2.html>` and
    :dcm:`Annex A.4 <part05/sect_A.4.html>`
    """
    frame_length = len(frame)
    # Add 1 to fix odd length frames not being caught
    if nr_fragments > (frame_length + 1) / 2.0:
        raise ValueError(
            "Too many fragments requested (the minimum fragment size is 2 bytes)"
        )

    length = int(frame_length / nr_fragments)

    # Each item shall be an even number of bytes
    if length % 2:
        length += 1

    # 1st to (N-1)th fragment
    for offset in range(0, length * (nr_fragments - 1), length):
        yield frame[offset : offset + length]

    # Nth fragment
    offset = length * (nr_fragments - 1)
    fragment = frame[offset:]

    # Pad last fragment if needed to make it even
    if (frame_length - offset) % 2:
        fragment += b"\x00"

    yield fragment


def itemize_fragment(fragment: bytes) -> bytes:
    """Return an itemized `fragment`.

    .. versionadded:: 1.2

    Parameters
    ----------
    fragment : bytes
        The fragment to itemize.

    Returns
    -------
    bytes
        The itemized fragment.

    Notes
    -----

    * The encoding of the item shall be in Little Endian.
    * Each fragment is encapsulated as a DICOM Item with tag (FFFE,E000), then
      a 4 byte length.
    """
    # item tag (FFFE,E000)
    item = b"\xFE\xFF\x00\xE0"
    # fragment length '<I' little endian, 4 byte unsigned int
    item += pack("<I", len(fragment))
    # fragment data
    item += fragment

    return item


def itemize_frame(frame: bytes, nr_fragments: int = 1) -> Iterator[bytes]:
    """Yield items generated from `frame`.

    .. versionadded:: 1.2

    Parameters
    ----------
    frame : bytes
        The data to fragment and itemise.
    nr_fragments : int, optional
        The number of fragments/items (default 1).

    Yields
    ------
    bytes
        An itemized fragment of the frame, encoded as little endian.

    Notes
    -----

    * The encoding of the items shall be in Little Endian.
    * Each fragment is encapsulated as a DICOM Item with tag (FFFE,E000), then
      a 4 byte length.

    References
    ----------
    DICOM Standard, Part 5, :dcm:`Section 7.5 <part05/sect_7.5.html>` and
    :dcm:`Annex A.4 <part05/sect_A.4.html>`
    """
    for fragment in fragment_frame(frame, nr_fragments):
        yield itemize_fragment(fragment)


def encapsulate(
    frames: list[bytes], fragments_per_frame: int = 1, has_bot: bool = True
) -> bytes:
    """Return encapsulated `frames`.

    .. versionadded:: 1.2

    When using a compressed transfer syntax (such as RLE Lossless or one of
    JPEG formats) then any *Pixel Data* must be :dcm:`encapsulated
    <part05/sect_A.4.html>`::

      # Where `frame1`, `frame2` are single frames that have been encoded
      # using the corresponding compression method to Transfer Syntax UID
      ds.PixelData = encapsulate([frame1, frame2, ...])

    For multi-frame data each frame must be encoded separately and then all
    encoded frames encapsulated together.

    When many large frames are to be encapsulated, the total length of
    encapsulated data may exceed the maximum length available with the
    :dcm:`Basic Offset Table<part05/sect_A.4.html>` (2**31 - 1 bytes). Under
    these circumstances you can:

    * Pass ``has_bot=False`` to :func:`~pydicom.encaps.encapsulate`
    * Use :func:`~pydicom.encaps.encapsulate_extended` and add the
      :dcm:`Extended Offset Table<part03/sect_C.7.6.3.html>` elements to your
      dataset (recommended)

    Data will be encapsulated with a Basic Offset Table Item at the beginning,
    then one or more fragment items. Each item will be of even length and the
    final fragment of each frame may be padded with ``0x00`` if required.

    Parameters
    ----------
    frames : list of bytes
        The frame data to encapsulate, one frame per item.
    fragments_per_frame : int, optional
        The number of fragments to use for each frame (default ``1``).
    has_bot : bool, optional
        ``True`` to include values in the Basic Offset Table, ``False``
        otherwise (default ``True``). If `fragments_per_frame` is not ``1``
        then it's strongly recommended that this be ``True``.

    Returns
    -------
    bytes
        The encapsulated pixel data.

    References
    ----------
    DICOM Standard, Part 5, :dcm:`Section 7.5 <part05/sect_7.5.html>` and
    :dcm:`Annex A.4 <part05/sect_A.4.html>`

    See Also
    --------
    :func:`~pydicom.encaps.encapsulate_buffer`
    :func:`~pydicom.encaps.encapsulate_extended`
    :func:`~pydicom.encaps.encapsulate_extended_buffer`
    """
    nr_frames = len(frames)
    output = bytearray()

    # Add the Basic Offset Table Item
    # Add the tag
    output.extend(b"\xFE\xFF\x00\xE0")
    if has_bot:
        # Check that the 2**32 - 1 limit in BOT item lengths won't be exceeded
        total = (nr_frames - 1) * 8 + sum([len(f) for f in frames[:-1]])
        if total > 2**32 - 1:
            raise ValueError(
                f"The total length of the encapsulated frame data ({total} "
                "bytes) will be greater than the maximum allowed by the Basic "
                f"Offset Table ({2**32 - 1} bytes), it's recommended that you "
                "use the Extended Offset Table instead (see the "
                "'encapsulate_extended' function for more information)"
            )

        # Add the length
        output.extend(pack("<I", 4 * nr_frames))
        # Reserve 4 x len(frames) bytes for the offsets
        output.extend(b"\xFF\xFF\xFF\xFF" * nr_frames)
    else:
        # Add the length
        output.extend(pack("<I", 0))

    bot_offsets = [0]
    for ii, frame in enumerate(frames):
        # `itemised_length` is the total length of each itemised frame
        itemised_length = 0
        for item in itemize_frame(frame, fragments_per_frame):
            itemised_length += len(item)
            output.extend(item)

        # Update the list of frame offsets
        bot_offsets.append(bot_offsets[ii] + itemised_length)

    if has_bot:
        # Go back and write the frame offsets - don't need the last offset
        output[8 : 8 + 4 * nr_frames] = pack(f"<{nr_frames}I", *bot_offsets[:-1])

    return bytes(output)


def encapsulate_buffer(
    buffers: list[BufferedIOBase], has_bot: bool = True
) -> EncapsulatedBuffer:
    """Return an :class:`~pydicom.encaps.EncapsulatedBuffer` instance from `buffers`.

    .. versionadded:: 3.0

    Examples
    --------

    .. code-block:: python

        from pydicom import Dataset, FileMetaDataset
        from pydicom.encaps import encapsulate_buffer
        from pydicom.uid import JPEG2000Lossless

        # Open the compressed image frames as io.BufferedReader instances
        frame1 = open("frame1.j2k", "rb")
        frame2 = open("frame2.j2k", "rb")
        frame3 = open("frame3.j2k", "rb")

        ds = Dataset()
        ds.file_meta = FileMetaDataset()
        ds.file_meta.TransferSyntaxUID = JPEG2000Lossless

        ds.PixelData = encapsulate_buffer([frame1, frame2, frame3])

        # Write the encapsulated buffer data to file
        ds.save_as("buffered_dataset.dcm")

        # Close the buffers
        frame1.close()
        frame2.close()
        frame3.close()

    Parameters
    ----------
    buffers : list[io.BufferedIOBase]
        A list of objects inheriting :class:`io.BufferedIOBase` containing the
        compressed *Pixel Data* frames to be encapsulated.

    Returns
    -------
    EncapsulatedBuffer
        A :class:`~pydicom.encaps.EncapsulatedBuffer` instance that can be used as
        the value for a *Pixel Data* element.

    See Also
    --------
    :func:`~pydicom.encaps.encapsulate`
    :func:`~pydicom.encaps.encapsulate_extended`
    :func:`~pydicom.encaps.encapsulate_extended_buffer`
    """
    return EncapsulatedBuffer(buffers, use_bot=has_bot)


def encapsulate_extended(frames: list[bytes]) -> tuple[bytes, bytes, bytes]:
    """Return encapsulated image data and values for the Extended Offset Table
    elements.

    When using a compressed transfer syntax (such as RLE Lossless or one of
    JPEG formats) then any *Pixel Data* must be :dcm:`encapsulated
    <part05/sect_A.4.html>`. When many large frames are to be encapsulated, the
    total length of encapsulated data may exceed the maximum offset available
    with the :dcm:`Basic Offset Table<part05/sect_A.4.html>` (2**32 - 1 bytes).
    Under these circumstances you can:

    * Use :func:`~pydicom.encaps.encapsulate_extended` and add the
      :dcm:`Extended Offset Table<part03/sect_C.7.6.3.html>` elements to your
      dataset (recommended)
    * Pass ``has_bot=False`` to :func:`~pydicom.encaps.encapsulate`

    Examples
    --------

    .. code-block:: python

        from pydicom import Dataset, FileMetaDataset
        from pydicom.encaps import encapsulate_extended
        from pydicom.uid import JPEG2000Lossless

        # 'frames' is a list of image frames that have been each been encoded
        # separately using the compression method corresponding to the Transfer
        # Syntax UID
        frames: list[bytes] = [...]
        out: tuple[bytes, bytes, bytes] = encapsulate_extended(frames)

        ds = Dataset()
        ds.file_meta = FileMetaDataset()
        ds.file_meta.TransferSyntaxUID = JPEG2000Lossless

        ds.PixelData = out[0]
        ds.ExtendedOffsetTable = out[1]
        ds.ExtendedOffsetTableLengths = out[2]

    Parameters
    ----------
    frames : list of bytes
        The compressed frame data to encapsulate, one frame per item.

    Returns
    -------
    bytes, bytes, bytes
        The (encapsulated frames, extended offset table, extended offset
        table lengths).

    See Also
    --------
    :func:`~pydicom.encaps.encapsulate`
    :func:`~pydicom.encaps.encapsulate_buffer`
    :func:`~pydicom.encaps.encapsulate_extended_buffer`
    """
    nr_frames = len(frames)
    frame_lengths = [len(frame) for frame in frames]
    # Odd-length frames get padded to even length by `encapsulate()`
    frame_lengths = [ii + ii % 2 for ii in frame_lengths]
    frame_offsets = [0]
    for ii, length in enumerate(frame_lengths[:-1]):
        # Extra 8 bytes for the Item tag and length
        frame_offsets.append(frame_offsets[ii] + length + 8)

    offsets = pack(f"<{nr_frames}Q", *frame_offsets)
    lengths = pack(f"<{nr_frames}Q", *frame_lengths)

    return encapsulate(frames, has_bot=False), offsets, lengths


def encapsulate_extended_buffer(
    buffers: list[BufferedIOBase],
) -> tuple[EncapsulatedBuffer, bytes, bytes]:
    """Return :class:`~pydicom.encaps.EncapsulatedBuffer` as well as encoded offsets
    and lengths for the Extended Offset Table elements.

    .. versionadded:: 3.0

    Examples
    --------

    .. code-block:: python

        from pydicom import Dataset, FileMetaDataset
        from pydicom.encaps import encapsulate_extended_buffer
        from pydicom.uid import JPEG2000Lossless

        # Open the compressed image frames as io.BufferedReader instances
        frame1 = open("frame1.j2k", "rb")
        frame2 = open("frame2.j2k", "rb")
        frame3 = open("frame3.j2k", "rb")

        out: tuple[EncapsulatedBuffer, bytes, bytes] = (
            encapsulate_extended_buffer([frame1, frame2, frame3])
        )

        ds = Dataset()
        ds.file_meta = FileMetaDataset()
        ds.file_meta.TransferSyntaxUID = JPEG2000Lossless

        ds.PixelData = out[0]
        ds.ExtendedOffsetTable = out[1]
        ds.ExtendedOffsetTableLengths = out[2]

        # Write the encapsulated buffer data to file
        ds.save_as("buffered_dataset.dcm")

        # Close the buffers
        frame1.close()
        frame2.close()
        frame3.close()

    Parameters
    ----------
    buffers : list[io.BufferedIOBase]
        A list of objects inheriting :class:`io.BufferedIOBase` containing the
        compressed *Pixel Data* frames to be encapsulated.

    Returns
    -------
    tuple[EncapsulatedBuffer, bytes, bytes]
        The (:class:`~pydicom.encaps.EncapsulatedBuffer`, extended offset table,
        extended offset table lengths).

    See Also
    --------
    :func:`~pydicom.encaps.encapsulate`
    :func:`~pydicom.encaps.encapsulate_buffer`
    :func:`~pydicom.encaps.encapsulate_extended`
    """
    eb = EncapsulatedBuffer(buffers)
    return eb, eb.extended_offsets, eb.extended_lengths


# Deprecated functions
def _get_frame_offsets(fp: DicomIO) -> tuple[bool, list[int]]:
    """Return a list of the fragment offsets from the Basic Offset Table.

    .. deprecated:: 3.0

        This function will be removed in v4.0, please use
        :func:`~pydicom.encaps.parse_basic_offsets` instead.

    **Basic Offset Table**

    The Basic Offset Table Item must be present and have a tag (FFFE,E000) and
    a length, however it may or may not have a value.

    Basic Offset Table with no value
    ::

        Item Tag   | Length    |
        FE FF 00 E0 00 00 00 00

    Basic Offset Table with value (2 frames)
    ::

        Item Tag   | Length    | Offset 1  | Offset 2  |
        FE FF 00 E0 08 00 00 00 00 00 00 00 10 00 00 00

    For single or multi-frame images with only one frame, the Basic Offset
    Table may or may not have a value. When it has no value then its length
    shall be ``0x00000000``.

    For multi-frame images with more than one frame, the Basic Offset Table
    should have a value containing concatenated 32-bit unsigned integer values
    that are the byte offsets to the first byte of the Item tag of the first
    fragment of each frame as measured from the first byte of the first item
    tag following the Basic Offset Table Item.

    All decoders, both for single and multi-frame images should accept both
    an empty Basic Offset Table and one containing offset values.

    .. versionchanged:: 1.4

        Changed to return (is BOT empty, list of offsets).

    Parameters
    ----------
    fp : filebase.DicomIO
        The encapsulated pixel data positioned at the start of the Basic Offset
        Table. ``fp.is_little_endian`` should be set to ``True``.

    Returns
    -------
    bool, list of int
        Whether or not the BOT is empty, and a list of the byte offsets
        to the first fragment of each frame, as measured from the start of the
        first item following the Basic Offset Table item.

    Raises
    ------
    ValueError
        If the Basic Offset Table item's tag is not (FFEE,E000) or if the
        length in bytes of the item's value is not a multiple of 4.

    References
    ----------
    DICOM Standard, Part 5, :dcm:`Annex A.4 <part05/sect_A.4.html>`
    """
    if not fp.is_little_endian:
        raise ValueError("'fp.is_little_endian' must be True")

    tag = Tag(fp.read_tag())

    if tag != 0xFFFEE000:
        raise ValueError(
            f"Unexpected tag '{tag}' when parsing the Basic Offset Table item"
        )

    length = fp.read_UL()
    if length % 4:
        raise ValueError(
            "The length of the Basic Offset Table item is not a multiple of 4"
        )

    offsets = []
    # Always return at least a 0 offset
    if length == 0:
        offsets.append(0)

    offsets.extend(fp.read_UL() for ii in range(length // 4))

    return bool(length), offsets


def _get_nr_fragments(fp: DicomIO) -> int:
    """Return the number of fragments in `fp`.

    .. deprecated:: 3.0

        This function will be removed in v4.0, please use
        :func:`~pydicom.encaps.parse_fragments` instead.
    """
    if not fp.is_little_endian:
        raise ValueError("'fp.is_little_endian' must be True")

    return parse_fragments(fp)[0]


def _generate_pixel_data_fragment(fp: DicomIO) -> Iterator[bytes]:
    """Yield the encapsulated pixel data fragments.

    .. deprecated:: 3.0

        This function will be remove in v4.0, please use
        :func:`~pydicom.encaps.generate_fragments` instead.

    For compressed (encapsulated) Transfer Syntaxes, the (7FE0,0010) *Pixel
    Data* element is encoded in an encapsulated format.

    **Encapsulation**

    The encoded pixel data stream is fragmented into one or more Items. The
    stream may represent a single or multi-frame image.

    Each *Data Stream Fragment* shall have tag of (FFFE,E000), followed by a 4
    byte *Item Length* field encoding the explicit number of bytes in the Item.
    All Items containing an encoded fragment shall have an even number of bytes
    greater than or equal to 2, with the last fragment being padded if
    necessary.

    The first Item in the Sequence of Items shall be a 'Basic Offset Table',
    however the Basic Offset Table item value is not required to be present.
    It is assumed that the Basic Offset Table item has already been read prior
    to calling this function (and that `fp` is positioned past this item).

    The remaining items in the Sequence of Items are the pixel data fragments
    and it is these items that will be read and returned by this function.

    The Sequence of Items is terminated by a (FFFE,E0DD) *Sequence Delimiter
    Item* with an Item Length field of value ``0x00000000``. The presence
    or absence of the *Sequence Delimiter Item* in `fp` has no effect on the
    returned fragments.

    *Encoding*

    The encoding of the data shall be little endian.

    Parameters
    ----------
    fp : filebase.DicomIO
        The encoded (7FE0,0010) *Pixel Data* element value, positioned at the
        start of the item tag for the first item after the Basic Offset Table
        item. ``fp.is_little_endian`` should be set to ``True``.

    Yields
    ------
    bytes
        A pixel data fragment.

    Raises
    ------
    ValueError
        If the data contains an item with an undefined length or an unknown
        tag.

    References
    ----------
    DICOM Standard Part 5, :dcm:`Annex A.4 <part05/sect_A.4.html>`
    """
    if not fp.is_little_endian:
        raise ValueError("'fp.is_little_endian' must be True")

    yield from generate_fragments(fp)


def _generate_pixel_data_frame(
    bytestream: bytes, nr_frames: int | None = None
) -> Iterator[bytes]:
    """Yield complete frames from `buffer` as :class:`bytes`.

    .. deprecated:: 3.0

        This function will be remove in v4.0, please use
        :func:`~pydicom.encaps.generate_frames` instead

    Parameters
    ----------
    bytestream : bytes
        The value of the (7FE0,0010) *Pixel Data* element from an encapsulated
        dataset. The Basic Offset Table item should be present and the
        Sequence Delimiter item may or may not be present.
    nr_frames : int, optional
        Required for multi-frame data when the Basic Offset Table is empty
        and there are multiple frames. This should be the value of (0028,0008)
        *Number of Frames*.

    Yields
    ------
    bytes
        A frame contained in the encapsulated pixel data.

    References
    ----------
    DICOM Standard Part 5, :dcm:`Annex A <part05/chapter_A.html>`
    """
    for frame in generate_fragmented_frames(bytestream, number_of_frames=nr_frames):
        yield b"".join(frame)


def _generate_pixel_data(
    bytestream: bytes, nr_frames: int | None = None
) -> Iterator[tuple[bytes, ...]]:
    """Yield an encapsulated pixel data frame.

    .. deprecated:: 3.0

        Please use :func:`~pydicom.encaps.generate_fragmented_frames` instead.

    For the following transfer syntaxes, a fragment may not contain encoded
    data from more than one frame. However data from one frame may span
    multiple fragments.

    * 1.2.840.10008.1.2.4.50 - JPEG Baseline (Process 1)
    * 1.2.840.10008.1.2.4.51 - JPEG Baseline (Process 2 and 4)
    * 1.2.840.10008.1.2.4.57 - JPEG Lossless, Non-Hierarchical (Process 14)
    * 1.2.840.10008.1.2.4.70 - JPEG Lossless, Non-Hierarchical, First-Order
      Prediction (Process 14 [Selection Value 1])
    * 1.2.840.10008.1.2.4.80 - JPEG-LS Lossless Image Compression
    * 1.2.840.10008.1.2.4.81 - JPEG-LS Lossy (Near-Lossless) Image Compression
    * 1.2.840.10008.1.2.4.90 - JPEG 2000 Image Compression (Lossless Only)
    * 1.2.840.10008.1.2.4.91 - JPEG 2000 Image Compression
    * 1.2.840.10008.1.2.4.92 - JPEG 2000 Part 2 Multi-component Image
      Compression (Lossless Only)
    * 1.2.840.10008.1.2.4.93 - JPEG 2000 Part 2 Multi-component Image
      Compression

    For the following transfer syntaxes, each frame shall be encoded in one and
    only one fragment.

    * 1.2.840.10008.1.2.5 - RLE Lossless

    Parameters
    ----------
    bytestream : bytes
        The value of the (7FE0,0010) *Pixel Data* element from an encapsulated
        dataset. The Basic Offset Table item should be present and the
        Sequence Delimiter item may or may not be present.
    nr_frames : int, optional
        Required for multi-frame data when the Basic Offset Table is empty
        and there are multiple frames. This should be the value of (0028,0008)
        *Number of Frames*.

    Yields
    -------
    tuple of bytes
        An encapsulated pixel data frame, with the contents of the
        :class:`tuple` the frame's fragmented data.

    Notes
    -----
    If the Basic Offset Table is empty and there are multiple fragments per
    frame then an attempt will be made to locate the frame boundaries by
    searching for the JPEG/JPEG-LS/JPEG2000 EOI/EOC marker (``0xFFD9``). If the
    marker is not present or the pixel data hasn't been compressed using one of
    the JPEG standards then the generated pixel data may be incorrect.

    References
    ----------
    DICOM Standard Part 5, :dcm:`Annex A <part05/chapter_A.html>`
    """
    yield from generate_fragmented_frames(bytestream, number_of_frames=nr_frames)


def _decode_data_sequence(data: bytes) -> list[bytes]:
    """Read encapsulated data and return a list of bytes.

    .. deprecated:: 3.0

        This function will be removed in v4.0, Please use
        :func:`~pydicom.encaps.generate_frames` for generating frame
        data or :func:`~pydicom.encaps.generate_fragments` for generating
        fragment data.

    Parameters
    ----------
    data : bytes
        The encapsulated data, typically the value from ``Dataset.PixelData``.

    Returns
    -------
    list of bytes
        All fragments as a list of ``bytes``.
    """
    # Convert data into a memory-mapped file
    with DicomBytesIO(data) as fp:
        # DICOM standard requires this
        fp.is_little_endian = True
        BasicOffsetTable = _read_item(fp)  # NOQA
        seq = []

        while True:
            item = _read_item(fp)

            # None is returned if get to Sequence Delimiter
            if not item:
                break
            seq.append(item)

        # XXX should
        return seq


def _defragment_data(data: bytes) -> bytes:
    """Read encapsulated data and return the fragments as one continuous bytes.

    .. deprecated:: 3.0

        This function will be removed in v4.0, Please use
        :func:`~pydicom.encaps.generate_frames` for generating frame
        data or :func:`~pydicom.encaps.generate_fragments` for generating
        fragment data.

    Parameters
    ----------
    data : bytes
        The encapsulated pixel data fragments.

    Returns
    -------
    bytes
        All fragments concatenated together.
    """
    return b"".join(_decode_data_sequence(data))


def _read_item(fp: DicomIO) -> bytes | None:
    """Read and return a single Item in the fragmented data stream.

    .. deprecated:: 3.0

        This function will be removed in v4.0, please use
        :func:`~pydicom.encaps.generate_fragments` instead.

    Parameters
    ----------
    fp : filebase.DicomIO
        The file-like to read the item from.

    Returns
    -------
    bytes
        The Item's raw bytes.
    """

    logger = config.logger
    try:
        tag = fp.read_tag()

    # already read delimiter before passing data here
    # so should just run out
    except EOFError:
        return None

    # No more items, time for sequence to stop reading
    if tag == SequenceDelimiterTag:
        length = fp.read_UL()
        logger.debug("%04x: Sequence Delimiter, length 0x%x", fp.tell() - 8, length)

        if length != 0:
            logger.warning(
                "Expected 0x00000000 after delimiter, found 0x%x,"
                " at data position 0x%x",
                length,
                fp.tell() - 4,
            )
        return None

    if tag != ItemTag:
        logger.warning(
            "Expected Item with tag %s at data position 0x%x", ItemTag, fp.tell() - 4
        )
        length = fp.read_UL()
    else:
        length = fp.read_UL()
        logger.debug("%04x: Item, length 0x%x", fp.tell() - 8, length)

    if length == 0xFFFFFFFF:
        raise ValueError(
            "Encapsulated data fragment had Undefined Length"
            f" at data position 0x{fp.tell() - 4:x}"
        )

    item_data = fp.read(length)
    return item_data


_DEPRECATED = {
    "get_frame_offsets": _get_frame_offsets,
    "get_nr_fragments": _get_nr_fragments,
    "generate_pixel_data_fragment": _generate_pixel_data_fragment,
    "generate_pixel_data_frame": _generate_pixel_data_frame,
    "generate_pixel_data": _generate_pixel_data,
    "decode_data_sequence": _decode_data_sequence,
    "defragment_data": _defragment_data,
    "read_item": _read_item,
    "itemise_frame": itemize_frame,
    "itemise_fragment": itemize_fragment,
}


def __getattr__(name: str) -> Any:
    if name in _DEPRECATED and not config._use_future:
        warn_and_log(
            f"{name} is deprecated and will be removed in v4.0",
            DeprecationWarning,
        )
        return _DEPRECATED[name]

    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



# TODO: remove module in v4.0
from typing import Any

from pydicom import config
from pydicom.misc import warn_and_log
from pydicom.pixels import get_encoder as _get_encoder
from pydicom.pixels.encoders import RLELosslessEncoder as _rle_encoder


_DEPRECATED = {
    "get_encoder": _get_encoder,
    "RLELosslessEncoder": _rle_encoder,
}


def __getattr__(name: str) -> Any:
    if name in _DEPRECATED and not config._use_future:
        msg = (
            f"The 'pydicom.encoders' module will be removed in v4.0, please use "
            f"'from pydicom.pixels import {name}' instead"
        )
        warn_and_log(msg, DeprecationWarning)
        return _DEPRECATED[name]

    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



# Copyright 2008-2024 pydicom authors. See LICENSE file for details.
"""Pixel data processing functions."""

from io import BytesIO
from struct import unpack, unpack_from
from typing import TYPE_CHECKING, cast

try:
    import numpy as np

    HAVE_NP = True
except ImportError:
    HAVE_NP = False

try:
    import PIL
    from PIL import ImageCms

    HAVE_PIL = True
except ImportError:
    HAVE_PIL = False

from pydicom.data import get_palette_files
from pydicom.misc import warn_and_log
from pydicom.uid import UID
from pydicom.valuerep import VR

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import Iterable
    from pydicom.dataset import Dataset


_CMS_COLOR_SPACES = {"SRGB": "sRGB", "XYZ": "XYZ", "LAB": "LAB"}
_CMS_INTENTS = {
    0: "procedural",
    1: "relative colorimetric",
    2: "saturation",
    3: "absolute colorimetric",
}


def apply_color_lut(
    arr: "np.ndarray", ds: "Dataset | None" = None, palette: str | UID | None = None
) -> "np.ndarray":
    """Apply a color palette lookup table to `arr`.

    If (0028,1201-1203) *Palette Color Lookup Table Data* are missing
    then (0028,1221-1223) *Segmented Palette Color Lookup Table Data* must be
    present and vice versa. The presence of (0028,1204) *Alpha Palette Color
    Lookup Table Data* or (0028,1224) *Alpha Segmented Palette Color Lookup
    Table Data* is optional.

    Use of this function with the :dcm:`Enhanced Palette Color Lookup Table
    Module<part03/sect_C.7.6.23.html>` or :dcm:`Supplemental Palette Color LUT
    Module<part03/sect_C.7.6.19.html>` is not currently supported.

    Parameters
    ----------
    arr : numpy.ndarray
        The pixel data to apply the color palette to.
    ds : dataset.Dataset, optional
        Required if `palette` is not supplied. A
        :class:`~pydicom.dataset.Dataset` containing a suitable
        :dcm:`Image Pixel<part03/sect_C.7.6.3.html>` or
        :dcm:`Palette Color Lookup Table<part03/sect_C.7.9.html>` Module.
    palette : str or uid.UID, optional
        Required if `ds` is not supplied. The name of one of the
        :dcm:`well-known<part06/chapter_B.html>` color palettes defined by the
        DICOM Standard. One of: ``'HOT_IRON'``, ``'PET'``,
        ``'HOT_METAL_BLUE'``, ``'PET_20_STEP'``, ``'SPRING'``, ``'SUMMER'``,
        ``'FALL'``, ``'WINTER'`` or the corresponding well-known (0008,0018)
        *SOP Instance UID*.

    Returns
    -------
    numpy.ndarray
        The RGB or RGBA pixel data as an array of ``np.uint8`` or ``np.uint16``
        values, depending on the 3rd value of (0028,1201) *Red Palette Color
        Lookup Table Descriptor*.

    References
    ----------

    * :dcm:`Image Pixel Module<part03/sect_C.7.6.3.html>`
    * :dcm:`Supplemental Palette Color LUT Module<part03/sect_C.7.6.19.html>`
    * :dcm:`Enhanced Palette Color LUT Module<part03/sect_C.7.6.23.html>`
    * :dcm:`Palette Colour LUT Module<part03/sect_C.7.9.html>`
    * :dcm:`Supplemental Palette Color LUTs
      <part03/sect_C.8.16.2.html#sect_C.8.16.2.1.1.1>`
    """
    # Note: input value (IV) is the stored pixel value in `arr`
    # LUTs[IV] -> [R, G, B] values at the IV pixel location in `arr`
    if not ds and not palette:
        raise ValueError("Either 'ds' or 'palette' is required")

    if palette:
        # Well-known palettes are all 8-bits per entry
        datasets = {
            "1.2.840.10008.1.5.1": "hotiron.dcm",
            "1.2.840.10008.1.5.2": "pet.dcm",
            "1.2.840.10008.1.5.3": "hotmetalblue.dcm",
            "1.2.840.10008.1.5.4": "pet20step.dcm",
            "1.2.840.10008.1.5.5": "spring.dcm",
            "1.2.840.10008.1.5.6": "summer.dcm",
            "1.2.840.10008.1.5.7": "fall.dcm",
            "1.2.840.10008.1.5.8": "winter.dcm",
        }
        if not UID(palette).is_valid:
            try:
                uids = {
                    "HOT_IRON": "1.2.840.10008.1.5.1",
                    "PET": "1.2.840.10008.1.5.2",
                    "HOT_METAL_BLUE": "1.2.840.10008.1.5.3",
                    "PET_20_STEP": "1.2.840.10008.1.5.4",
                    "SPRING": "1.2.840.10008.1.5.5",
                    "SUMMER": "1.2.840.10008.1.5.6",
                    "FALL": "1.2.840.10008.1.5.8",
                    "WINTER": "1.2.840.10008.1.5.7",
                }
                palette = uids[palette]
            except KeyError:
                raise ValueError(f"Unknown palette '{palette}'")

        try:
            from pydicom import dcmread

            fname = datasets[palette]
            ds = dcmread(get_palette_files(fname)[0])
        except KeyError:
            raise ValueError(f"Unknown palette '{palette}'")

    ds = cast("Dataset", ds)

    # C.8.16.2.1.1.1: Supplemental Palette Color LUT
    # TODO: Requires greyscale visualisation pipeline
    if getattr(ds, "PixelPresentation", None) in ["MIXED", "COLOR"]:
        raise ValueError(
            "Use of this function with the Supplemental Palette Color Lookup "
            "Table Module is not currently supported"
        )

    if "RedPaletteColorLookupTableDescriptor" not in ds:
        raise ValueError("No suitable Palette Color Lookup Table Module found")

    # All channels are supposed to be identical
    lut_desc = cast(list[int], ds.RedPaletteColorLookupTableDescriptor)
    # A value of 0 = 2^16 entries
    nr_entries = lut_desc[0] or 2**16

    # May be negative if Pixel Representation is 1
    first_map = lut_desc[1]
    # Actual bit depth may be larger (8 bit entries in 16 bits allocated)
    nominal_depth = lut_desc[2]
    dtype = np.dtype(f"uint{nominal_depth:.0f}")

    luts: list[bytes] = []
    if "RedPaletteColorLookupTableData" in ds:
        # LUT Data is described by PS3.3, C.7.6.3.1.6
        r_lut = cast(bytes, ds.RedPaletteColorLookupTableData)
        g_lut = cast(bytes, ds.GreenPaletteColorLookupTableData)
        b_lut = cast(bytes, ds.BluePaletteColorLookupTableData)
        a_lut = cast(
            bytes | None, getattr(ds, "AlphaPaletteColorLookupTableData", None)
        )

        actual_depth = len(r_lut) / nr_entries * 8
        dtype = np.dtype(f"uint{actual_depth:.0f}")

        luts.extend(
            np.frombuffer(lut_bytes, dtype=dtype)
            for lut_bytes in (r_lut, g_lut, b_lut, a_lut)
            if lut_bytes
        )
    elif "SegmentedRedPaletteColorLookupTableData" in ds:
        # Segmented LUT Data is described by PS3.3, C.7.9.2
        r_lut = cast(bytes, ds.SegmentedRedPaletteColorLookupTableData)
        g_lut = cast(bytes, ds.SegmentedGreenPaletteColorLookupTableData)
        b_lut = cast(bytes, ds.SegmentedBluePaletteColorLookupTableData)
        a_lut = cast(
            bytes | None, getattr(ds, "SegmentedAlphaPaletteColorLookupTableData", None)
        )

        if hasattr(ds, "file_meta"):
            is_little_endian = ds.file_meta._tsyntax_encoding[1]
        else:
            is_little_endian = ds.original_encoding[1]

        if is_little_endian is None:
            raise AttributeError(
                "Unable to determine the endianness of the dataset, please set "
                "an appropriate Transfer Syntax UID in "
                f"'{type(ds).__name__}.file_meta'"
            )

        endianness = "><"[is_little_endian]
        byte_depth = nominal_depth // 8
        fmt = "B" if byte_depth == 1 else "H"
        actual_depth = nominal_depth

        for seg in (r_lut, g_lut, b_lut, a_lut):
            if seg:
                len_seg = len(seg) // byte_depth
                s_fmt = f"{endianness}{len_seg}{fmt}"
                lut_ints = _expand_segmented_lut(unpack(s_fmt, seg), s_fmt)
                luts.append(np.asarray(lut_ints, dtype=dtype))
    else:
        raise ValueError("No suitable Palette Color Lookup Table Module found")

    if actual_depth not in [8, 16]:
        raise ValueError(
            f"The bit depth of the LUT data '{actual_depth:.1f}' "
            "is invalid (only 8 or 16 bits per entry allowed)"
        )

    lut_lengths = [len(ii) for ii in luts]
    if not all(ii == lut_lengths[0] for ii in lut_lengths[1:]):
        raise ValueError("LUT data must be the same length")

    # IVs < `first_map` get set to first LUT entry (i.e. index 0)
    clipped_iv = np.zeros(arr.shape, dtype=dtype)
    # IVs >= `first_map` are mapped by the Palette Color LUTs
    # `first_map` may be negative, positive or 0
    mapped_pixels = arr >= first_map
    clipped_iv[mapped_pixels] = arr[mapped_pixels] - np.int32(first_map)
    # IVs > number of entries get set to last entry
    np.clip(clipped_iv, 0, nr_entries - 1, out=clipped_iv)

    # Output array may be RGB or RGBA
    out = np.empty(list(arr.shape) + [len(luts)], dtype=dtype)
    for ii, lut in enumerate(luts):
        out[..., ii] = lut[clipped_iv]

    return out


def apply_icc_profile(
    arr: "np.ndarray",
    ds: "Dataset | None" = None,
    transform: "PIL.ImageCms.ImageCmsTransform | None" = None,
    intent: int | None = None,
    color_space: str | None = None,
) -> "np.ndarray":
    """Apply an `ICC Profile <https://www.color.org/iccprofile.xalter>`_ to `arr`,
    either from the dataset `ds` or an existing Pillow color transformation object
    `transform`.

    .. versionadded:: 3.0

    .. warning::

        This function requires `NumPy <https://numpy.org/>`_ and `Pillow
        <https://pillow.readthedocs.io/en/stable/>`_.


    Parameters
    ----------
    arr : numpy.ndarray
        8-bit RGB pixel data to apply an ICC profile to, shaped as either (rows,
        columns, samples) or (frames, rows, columns, samples).
    ds : pydicom.dataset.Dataset, optional
        Required if `transform` is not supplied, a :class:`~pydicom.dataset.Dataset`
        containing elements from the :dcm:`ICC Profile<part03/sect_C.11.15.html>`
        module.
    transform : PIL.ImageCms.ImageCmsTransform, optional
        An :class:`~PIL.ImageCms.ImageCmsTransform` instance such as is returned by
        the :func:`create_icc_transform` function. Required if `ds` is not used and
        recommended when the same ICC profile is to be re-used on multiple ndarrays.
    intent : int, optional
        If `transform` is not supplied this is the rendering intent of the
        transformation that will be created from `ds`, one of:

        * ``0``: perceptual
        * ``1``: relative colorimetric
        * ``2``: saturation
        * ``3``: absolute colorimetric

        If no `intent` is specified then the default rendering intent in the ICC
        profile will be used.
    color_space : str, optional
        If `transform` is not supplied this is the output color space to use
        for the transformation created from `ds`. If not used then defaults to the
        (0028,2002) *Color Space* value if its present in `ds`, otherwise defaults to
        ``"sRGB"``. If used then it will override the (0028,2002) *Color Space* element
        value (if present). Must be one of the :func:`supported Pillow color spaces
        <PIL.ImageCms.createProfile>`, although in practice only ``"sRGB"`` is likely
        to be used.

    Returns
    -------
    np.ndarray
        The input ndarray with the color transformation applied in-place.
    """
    if not HAVE_PIL:
        raise ImportError("Pillow is required to apply an ICC profile to an ndarray")

    # One of 'ds' or 'transform', but not both
    if ds is None and not transform:
        raise ValueError("Either 'ds' or 'transform' must be supplied")

    if ds is not None and transform:
        raise ValueError("Only one of 'ds' and 'transform' should be used, not both")

    if arr.ndim not in (3, 4):
        raise ValueError(f"The ndarray must have 3 or 4 dimensions, not {arr.ndim}")

    if arr.shape[-1] != 3:
        raise ValueError(
            "Invalid ndarray shape, must be (rows, columns, 3) or (frames, rows, "
            f"columns, 3), not {arr.shape}"
        )

    if not transform:
        transform = create_icc_transform(
            ds=ds,
            intent=intent,
            color_space=color_space,
        )

    if arr.ndim == 4:
        for idx, frame in enumerate(arr):
            im = PIL.Image.fromarray(frame)  # type: ignore[no-untyped-call]
            ImageCms.applyTransform(im, transform, inPlace=True)
            arr[idx] = np.array(im)

        return arr

    im = PIL.Image.fromarray(arr)  # type: ignore[no-untyped-call]
    ImageCms.applyTransform(im, transform, inPlace=True)

    arr[...] = np.array(im)

    return arr


def apply_modality_lut(arr: "np.ndarray", ds: "Dataset") -> "np.ndarray":
    """Apply a modality lookup table or rescale operation to `arr`.

    Parameters
    ----------
    arr : numpy.ndarray
        The :class:`~numpy.ndarray` to apply the modality LUT or rescale
        operation to.
    ds : dataset.Dataset
        A dataset containing a :dcm:`Modality LUT Module
        <part03/sect_C.11.html#sect_C.11.1>`.

    Returns
    -------
    numpy.ndarray
        An array with applied modality LUT or rescale operation. If
        (0028,3000) *Modality LUT Sequence* is present then returns an array
        of ``np.uint8`` or ``np.uint16``, depending on the 3rd value of
        (0028,3002) *LUT Descriptor*. If (0028,1052) *Rescale Intercept* and
        (0028,1053) *Rescale Slope* are present then returns an array of
        ``np.float64``. If neither are present then `arr` will be returned
        unchanged.

    Notes
    -----
    When *Rescale Slope* and *Rescale Intercept* are used, the output range
    is from (min. pixel value * Rescale Slope + Rescale Intercept) to
    (max. pixel value * Rescale Slope + Rescale Intercept), where min. and
    max. pixel value are determined from (0028,0101) *Bits Stored* and
    (0028,0103) *Pixel Representation*.

    References
    ----------
    * DICOM Standard, Part 3, :dcm:`Annex C.11.1
      <part03/sect_C.11.html#sect_C.11.1>`
    * DICOM Standard, Part 4, :dcm:`Annex N.2.1.1
      <part04/sect_N.2.html#sect_N.2.1.1>`
    """
    if ds.get("ModalityLUTSequence"):
        item = cast(list["Dataset"], ds.ModalityLUTSequence)[0]
        nr_entries = cast(list[int], item.LUTDescriptor)[0] or 2**16
        first_map = cast(list[int], item.LUTDescriptor)[1]
        nominal_depth = cast(list[int], item.LUTDescriptor)[2]

        dtype = f"uint{nominal_depth}"

        # Ambiguous VR, US or OW
        unc_data: Iterable[int]
        if item["LUTData"].VR == VR.OW:
            if hasattr(ds, "file_meta"):
                is_little_endian = ds.file_meta._tsyntax_encoding[1]
            else:
                is_little_endian = ds.original_encoding[1]

            if is_little_endian is None:
                raise AttributeError(
                    "Unable to determine the endianness of the dataset, please set "
                    "an appropriate Transfer Syntax UID in "
                    f"'{type(ds).__name__}.file_meta'"
                )

            endianness = "><"[is_little_endian]
            unpack_fmt = f"{endianness}{nr_entries}H"
            unc_data = unpack(unpack_fmt, cast(bytes, item.LUTData))
        else:
            unc_data = cast(list[int], item.LUTData)

        lut_data: np.ndarray = np.asarray(unc_data, dtype=dtype)

        # IVs < `first_map` get set to first LUT entry (i.e. index 0)
        clipped_iv = np.zeros(arr.shape, dtype=arr.dtype)
        # IVs >= `first_map` are mapped by the Modality LUT
        # `first_map` may be negative, positive or 0
        mapped_pixels = arr >= first_map
        clipped_iv[mapped_pixels] = arr[mapped_pixels] - first_map
        # IVs > number of entries get set to last entry
        np.clip(clipped_iv, 0, nr_entries - 1, out=clipped_iv)

        return lut_data[clipped_iv]
    elif "RescaleSlope" in ds and "RescaleIntercept" in ds:
        arr = arr.astype(np.float64) * cast(float, ds.RescaleSlope)
        arr += cast(float, ds.RescaleIntercept)

    return arr


def apply_presentation_lut(arr: "np.ndarray", ds: "Dataset") -> "np.ndarray":
    """Apply a Presentation LUT to `arr` and return the P-values.

    Parameters
    ----------
    arr : numpy.ndarray
        The :class:`~numpy.ndarray` to apply the presentation LUT operation to.
    ds : dataset.Dataset
        A dataset containing :dcm:`Presentation LUT Module
        <part03/sect_C.11.4.html>` elements.

    Returns
    -------
    numpy.ndarray
        If a Presentation LUT Module is present in `ds` then returns an array
        of P-values, otherwise returns `arr` unchanged.

    Notes
    -----
    If the dataset the *Pixel Data* originated from contains a Modality LUT
    and/or VOI LUT then they must be applied before the Presentation LUT.

    See Also
    --------
    :func:`~pydicom.pixels.processing.apply_modality_lut`
    :func:`~pydicom.pixels.processing.apply_voi_lut`
    """
    if "PresentationLUTSequence" in ds:
        item = ds.PresentationLUTSequence[0]
        # nr_entries is the number of entries in the LUT
        # first_map is the first input value mapped and shall always be 0
        # bit_depth is number of bits for each entry, up to 16
        nr_entries, first_map, bit_depth = item.LUTDescriptor
        nr_entries = 2**16 if nr_entries == 0 else nr_entries

        itemsize = 8 if bit_depth <= 8 else 16
        nr_bytes = nr_entries * (itemsize // 8)

        # P-values to be mapped to the input, always unsigned
        # LUTData is (US or OW)
        elem = item["LUTData"]
        if elem.VR == VR.US:
            lut = np.asarray(elem.value, dtype="u2")
        else:
            lut = np.frombuffer(item.LUTData[:nr_bytes], dtype=f"uint{itemsize}")

        # Set any unused bits to an appropriate value
        if bit_shift := itemsize - bit_depth:
            if not lut.flags.writeable:
                lut = lut.copy()

            np.left_shift(lut, bit_shift, out=lut)
            np.right_shift(lut, bit_shift, out=lut)

        # Linearly scale `arr` to quantize it to `nr_entries` values
        arr = arr.astype("float32")
        arr -= arr.min()
        arr /= arr.max() / (nr_entries - 1)
        arr = arr.astype("uint16")

        return cast("np.ndarray", lut[arr])

    if "PresentationLUTShape" in ds:
        transform = ds.PresentationLUTShape.strip().upper()
        if transform not in ("IDENTITY", "INVERSE"):
            raise NotImplementedError(
                "A (2050,0020) 'Presentation LUT Shape' value of "
                f"'{ds.PresentationLUTShape}' is not supported"
            )

        if transform == "INVERSE":
            arr = arr.max() - arr

    return arr


apply_rescale = apply_modality_lut


def apply_voi_lut(
    arr: "np.ndarray", ds: "Dataset", index: int = 0, prefer_lut: bool = True
) -> "np.ndarray":
    """Apply a VOI lookup table or windowing operation to `arr`.

    .. versionchanged:: 2.1

        Added the `prefer_lut` keyword parameter

    Parameters
    ----------
    arr : numpy.ndarray
        The :class:`~numpy.ndarray` to apply the VOI LUT or windowing operation
        to.
    ds : dataset.Dataset
        A dataset containing a :dcm:`VOI LUT Module<part03/sect_C.11.2.html>`.
        If (0028,3010) *VOI LUT Sequence* is present then returns an array
        of ``np.uint8`` or ``np.uint16``, depending on the 3rd value of
        (0028,3002) *LUT Descriptor*. If (0028,1050) *Window Center* and
        (0028,1051) *Window Width* are present then returns an array of
        ``np.float64``. If neither are present then `arr` will be returned
        unchanged.
    index : int, optional
        When the VOI LUT Module contains multiple alternative views, this is
        the index of the view to return (default ``0``).
    prefer_lut : bool
        When the VOI LUT Module contains both *Window Width*/*Window Center*
        and *VOI LUT Sequence*, if ``True`` (default) then apply the VOI LUT,
        otherwise apply the windowing operation.

    Returns
    -------
    numpy.ndarray
        An array with applied VOI LUT or windowing operation.

    Notes
    -----
    When the dataset requires a modality LUT or rescale operation as part of
    the Modality LUT module then that must be applied before any windowing
    operation.

    See Also
    --------
    :func:`~pydicom.pixels.processing.apply_modality_lut`
    :func:`~pydicom.pixels.processing.apply_voi`
    :func:`~pydicom.pixels.processing.apply_windowing`

    References
    ----------
    * DICOM Standard, Part 3, :dcm:`Annex C.11.2
      <part03/sect_C.11.html#sect_C.11.2>`
    * DICOM Standard, Part 3, :dcm:`Annex C.8.11.3.1.5
      <part03/sect_C.8.11.3.html#sect_C.8.11.3.1.5>`
    * DICOM Standard, Part 4, :dcm:`Annex N.2.1.1
      <part04/sect_N.2.html#sect_N.2.1.1>`
    """
    valid_voi = False
    if ds.get("VOILUTSequence"):
        ds.VOILUTSequence = cast(list["Dataset"], ds.VOILUTSequence)
        valid_voi = None not in [
            ds.VOILUTSequence[0].get("LUTDescriptor", None),
            ds.VOILUTSequence[0].get("LUTData", None),
        ]
    valid_windowing = None not in [
        ds.get("WindowCenter", None),
        ds.get("WindowWidth", None),
    ]

    if valid_voi and valid_windowing:
        if prefer_lut:
            return apply_voi(arr, ds, index)

        return apply_windowing(arr, ds, index)

    if valid_voi:
        return apply_voi(arr, ds, index)

    if valid_windowing:
        return apply_windowing(arr, ds, index)

    return arr


def apply_voi(arr: "np.ndarray", ds: "Dataset", index: int = 0) -> "np.ndarray":
    """Apply a VOI lookup table to `arr`.

    .. versionadded:: 2.1

    See :func:`~pydicom.pixels.processing.apply_voi_lut` for applying *Window
    Width*/*Window Center* as a fallback if no *VOI LUT Sequence* is present.

    Parameters
    ----------
    arr : numpy.ndarray
        The :class:`~numpy.ndarray` to apply the VOI LUT to.
    ds : dataset.Dataset
        A dataset containing a :dcm:`VOI LUT Module<part03/sect_C.11.2.html>`.
        If (0028,3010) *VOI LUT Sequence* is present then returns an array
        of ``np.uint8`` or ``np.uint16``, depending on the 3rd value of
        (0028,3002) *LUT Descriptor*, otherwise `arr` will be returned
        unchanged.
    index : int, optional
        When the VOI LUT Module contains multiple alternative views, this is
        the index of the view to return (default ``0``).

    Returns
    -------
    numpy.ndarray
        An array with applied VOI LUT.

    See Also
    --------
    :func:`~pydicom.pixels.processing.apply_modality_lut`
    :func:`~pydicom.pixels.processing.apply_windowing`
    :func:`~pydicom.pixels.processing.apply_voi_lut`

    References
    ----------
    * DICOM Standard, Part 3, :dcm:`Annex C.11.2
      <part03/sect_C.11.html#sect_C.11.2>`
    * DICOM Standard, Part 3, :dcm:`Annex C.8.11.3.1.5
      <part03/sect_C.8.11.3.html#sect_C.8.11.3.1.5>`
    * DICOM Standard, Part 4, :dcm:`Annex N.2.1.1
      <part04/sect_N.2.html#sect_N.2.1.1>`
    """
    if not ds.get("VOILUTSequence"):
        return arr

    if not np.issubdtype(arr.dtype, np.integer):
        warn_and_log(
            "Applying a VOI LUT on a float input array may give incorrect results"
        )

    # VOI LUT Sequence contains one or more items
    item = cast(list["Dataset"], ds.VOILUTSequence)[index]
    lut_descriptor = cast(list[int], item.LUTDescriptor)
    nr_entries = lut_descriptor[0] or 2**16
    first_map = lut_descriptor[1]

    # PS3.3 C.8.11.3.1.5: may be 8, 10-16
    nominal_depth = lut_descriptor[2]
    if nominal_depth in list(range(10, 17)):
        dtype = "uint16"
    elif nominal_depth == 8:
        dtype = "uint8"
    else:
        raise NotImplementedError(
            f"'{nominal_depth}' bits per LUT entry is not supported"
        )

    # Ambiguous VR, US or OW
    unc_data: Iterable[int]
    if item["LUTData"].VR == VR.OW:
        if hasattr(ds, "file_meta"):
            is_little_endian = ds.file_meta._tsyntax_encoding[1]
        else:
            is_little_endian = ds.original_encoding[1]

        if is_little_endian is None:
            raise AttributeError(
                "Unable to determine the endianness of the dataset, please set "
                "an appropriate Transfer Syntax UID in "
                f"'{type(ds).__name__}.file_meta'"
            )

        unpack_fmt = f"{'><'[is_little_endian]}{nr_entries}H"
        unc_data = unpack_from(unpack_fmt, cast(bytes, item.LUTData))
    else:
        unc_data = cast(list[int], item.LUTData)

    lut_data: np.ndarray = np.asarray(unc_data, dtype=dtype)

    # IVs < `first_map` get set to first LUT entry (i.e. index 0)
    clipped_iv = np.zeros(arr.shape, dtype=dtype)
    # IVs >= `first_map` are mapped by the VOI LUT
    # `first_map` may be negative, positive or 0
    mapped_pixels = arr >= first_map
    clipped_iv[mapped_pixels] = arr[mapped_pixels] - first_map
    # IVs > number of entries get set to last entry
    np.clip(clipped_iv, 0, nr_entries - 1, out=clipped_iv)

    return cast("np.ndarray", lut_data[clipped_iv])


def apply_windowing(arr: "np.ndarray", ds: "Dataset", index: int = 0) -> "np.ndarray":
    """Apply a windowing operation to `arr`.

    .. versionadded:: 2.1

    Parameters
    ----------
    arr : numpy.ndarray
        The :class:`~numpy.ndarray` to apply the windowing operation to.
    ds : dataset.Dataset
        A dataset containing a :dcm:`VOI LUT Module<part03/sect_C.11.2.html>`.
        If (0028,1050) *Window Center* and (0028,1051) *Window Width* are
        present then returns an array of ``np.float64``, otherwise `arr` will
        be returned unchanged.
    index : int, optional
        When the VOI LUT Module contains multiple alternative views, this is
        the index of the view to return (default ``0``).

    Returns
    -------
    numpy.ndarray
        An array with applied windowing operation.

    Notes
    -----
    When the dataset requires a modality LUT or rescale operation as part of
    the Modality LUT module then that must be applied before any windowing
    operation.

    See Also
    --------
    :func:`~pydicom.pixels.processing.apply_modality_lut`
    :func:`~pydicom.pixels.processing.apply_voi`
    :func:`~pydicom.pixels.processing.apply_voi_lut`

    References
    ----------
    * DICOM Standard, Part 3, :dcm:`Annex C.11.2
      <part03/sect_C.11.html#sect_C.11.2>`
    * DICOM Standard, Part 3, :dcm:`Annex C.8.11.3.1.5
      <part03/sect_C.8.11.3.html#sect_C.8.11.3.1.5>`
    * DICOM Standard, Part 4, :dcm:`Annex N.2.1.1
      <part04/sect_N.2.html#sect_N.2.1.1>`
    """
    if "WindowWidth" not in ds and "WindowCenter" not in ds:
        return arr

    if ds.PhotometricInterpretation not in ["MONOCHROME1", "MONOCHROME2"]:
        raise ValueError(
            "When performing a windowing operation only 'MONOCHROME1' and "
            "'MONOCHROME2' are allowed for (0028,0004) Photometric "
            "Interpretation"
        )

    # May be LINEAR (default), LINEAR_EXACT, SIGMOID or not present, VM 1
    voi_func = cast(str, getattr(ds, "VOILUTFunction", "LINEAR")).upper()
    # VR DS, VM 1-n
    elem = ds["WindowCenter"]
    center = cast(list[float], elem.value)[index] if elem.VM > 1 else elem.value
    center = cast(float, center)
    elem = ds["WindowWidth"]
    width = cast(list[float], elem.value)[index] if elem.VM > 1 else elem.value
    width = cast(float, width)

    # The output range depends on whether or not a modality LUT or rescale
    #   operation has been applied
    ds.BitsStored = cast(int, ds.BitsStored)
    y_min: float
    y_max: float
    if ds.get("ModalityLUTSequence"):
        # Unsigned - see PS3.3 C.11.1.1.1
        y_min = 0
        item = cast(list["Dataset"], ds.ModalityLUTSequence)[0]
        bit_depth = cast(list[int], item.LUTDescriptor)[2]
        y_max = 2**bit_depth - 1
    elif ds.PixelRepresentation == 0:
        # Unsigned
        y_min = 0
        y_max = 2**ds.BitsStored - 1
    else:
        # Signed
        y_min = -(2 ** (ds.BitsStored - 1))
        y_max = 2 ** (ds.BitsStored - 1) - 1

    slope = ds.get("RescaleSlope", None)
    intercept = ds.get("RescaleIntercept", None)
    if slope is not None and intercept is not None:
        ds.RescaleSlope = cast(float, ds.RescaleSlope)
        ds.RescaleIntercept = cast(float, ds.RescaleIntercept)
        # Otherwise its the actual data range
        y_min = y_min * ds.RescaleSlope + ds.RescaleIntercept
        y_max = y_max * ds.RescaleSlope + ds.RescaleIntercept

    y_range = y_max - y_min
    arr = arr.astype("float64")

    if voi_func in ["LINEAR", "LINEAR_EXACT"]:
        # PS3.3 C.11.2.1.2.1 and C.11.2.1.3.2
        if voi_func == "LINEAR":
            if width < 1:
                raise ValueError(
                    "The (0028,1051) Window Width must be greater than or "
                    "equal to 1 for a 'LINEAR' windowing operation"
                )
            center -= 0.5
            width -= 1
        elif width <= 0:
            raise ValueError(
                "The (0028,1051) Window Width must be greater than 0 "
                "for a 'LINEAR_EXACT' windowing operation"
            )

        below = arr <= (center - width / 2)
        above = arr > (center + width / 2)
        between = np.logical_and(~below, ~above)

        arr[below] = y_min
        arr[above] = y_max
        if between.any():
            arr[between] = ((arr[between] - center) / width + 0.5) * y_range + y_min
    elif voi_func == "SIGMOID":
        # PS3.3 C.11.2.1.3.1
        if width <= 0:
            raise ValueError(
                "The (0028,1051) Window Width must be greater than 0 "
                "for a 'SIGMOID' windowing operation"
            )

        arr = y_range / (1 + np.exp(-4 * (arr - center) / width)) + y_min
    else:
        raise ValueError(f"Unsupported (0028,1056) VOI LUT Function value '{voi_func}'")

    return arr


def convert_color_space(
    arr: "np.ndarray", current: str, desired: str, per_frame: bool = False
) -> "np.ndarray":
    """Convert the image(s) in `arr` from one color space to another.

    .. versionchanged:: 2.2

        Added `per_frame` keyword parameter.

    Parameters
    ----------
    arr : numpy.ndarray
        The image(s) as :class:`numpy.ndarray` with :attr:`~numpy.ndarray.shape`
        (frames, rows, columns, 3) or (rows, columns, 3) and a 'uint8'
        :class:`~numpy.dtype` (unsigned 8-bit).
    current : str
        The current color space, should be a valid value for (0028,0004)
        *Photometric Interpretation*. One of ``'RGB'``, ``'YBR_FULL'``,
        ``'YBR_FULL_422'``.
    desired : str
        The desired color space, should be a valid value for (0028,0004)
        *Photometric Interpretation*. One of ``'RGB'``, ``'YBR_FULL'``,
        ``'YBR_FULL_422'``.
    per_frame : bool, optional
        If ``True`` and the input array contains multiple frames then process
        each frame individually and update `arr` in-place to reduce memory
        usage. Default ``False``.

    Returns
    -------
    numpy.ndarray
        The image(s) converted to the desired color space. If `per_frame` is
        ``False`` (the default) then a new :class:`~numpy.ndarray` will be
        returned, otherwise `arr` will be updated in-place.

    References
    ----------

    * DICOM Standard, Part 3,
      :dcm:`Annex C.7.6.3.1.2<part03/sect_C.7.6.3.html#sect_C.7.6.3.1.2>`
    * ISO/IEC 10918-5:2012 (`ITU T.871
      <https://www.ijg.org/files/T-REC-T.871-201105-I!!PDF-E.pdf>`_),
      Section 7
    """
    if arr.dtype != np.dtype("u1"):
        raise ValueError(
            f"Invalid ndarray.dtype '{arr.dtype}' for color space conversion, "
            "must be 'uint8' or an equivalent"
        )

    def _no_change(arr: "np.ndarray") -> "np.ndarray":
        return arr

    _converters = {
        "YBR_FULL_422": {
            "YBR_FULL_422": _no_change,
            "YBR_FULL": _no_change,
            "RGB": _convert_YBR_FULL_to_RGB,
        },
        "YBR_FULL": {
            "YBR_FULL": _no_change,
            "YBR_FULL_422": _no_change,
            "RGB": _convert_YBR_FULL_to_RGB,
        },
        "RGB": {
            "RGB": _no_change,
            "YBR_FULL": _convert_RGB_to_YBR_FULL,
            "YBR_FULL_422": _convert_RGB_to_YBR_FULL,
        },
    }
    try:
        converter = _converters[current][desired]
    except KeyError:
        raise NotImplementedError(
            f"Conversion from {current} to {desired} is not supported."
        )

    if len(arr.shape) == 4 and per_frame:
        for idx, frame in enumerate(arr):
            arr[idx] = converter(frame)

        return arr

    return converter(arr)


def _convert_RGB_to_YBR_FULL(arr: "np.ndarray") -> "np.ndarray":
    """Return an ndarray converted from RGB to YBR_FULL color space.

    Parameters
    ----------
    arr : numpy.ndarray
        An ndarray of an 8-bit per channel images in RGB color space.

    Returns
    -------
    numpy.ndarray
        The array in YBR_FULL color space.

    References
    ----------

    * DICOM Standard, Part 3,
      :dcm:`Annex C.7.6.3.1.2<part03/sect_C.7.6.3.html#sect_C.7.6.3.1.2>`
    * ISO/IEC 10918-5:2012 (`ITU T.871
      <https://www.ijg.org/files/T-REC-T.871-201105-I!!PDF-E.pdf>`_),
      Section 7
    """
    orig_dtype = arr.dtype

    rgb_to_ybr = np.asarray(
        [
            [+0.299, -0.299 / 1.772, +0.701 / 1.402],
            [+0.587, -0.587 / 1.772, -0.587 / 1.402],
            [+0.114, +0.886 / 1.772, -0.114 / 1.402],
        ],
        dtype=np.float32,
    )

    arr = np.matmul(arr, rgb_to_ybr, dtype=np.float32)
    arr += [0.5, 128.5, 128.5]
    # Round(x) -> floor of (arr + 0.5) : 0.5 added in previous step
    np.floor(arr, out=arr)
    # Max(0, arr) -> 0 if 0 >= arr, arr otherwise
    # Min(arr, 255) -> arr if arr <= 255, 255 otherwise
    np.clip(arr, 0, 255, out=arr)

    return arr.astype(orig_dtype)


def _convert_YBR_FULL_to_RGB(arr: "np.ndarray") -> "np.ndarray":
    """Return an ndarray converted from YBR_FULL to RGB color space.

    Parameters
    ----------
    arr : numpy.ndarray
        An ndarray of an 8-bit per channel images in YBR_FULL color space.

    Returns
    -------
    numpy.ndarray
        The array in RGB color space.

    References
    ----------

    * DICOM Standard, Part 3,
      :dcm:`Annex C.7.6.3.1.2<part03/sect_C.7.6.3.html#sect_C.7.6.3.1.2>`
    * ISO/IEC 10918-5:2012, Section 7
    """
    orig_dtype = arr.dtype

    ybr_to_rgb = np.asarray(
        [
            [1.000, 1.000, 1.000],
            [0.000, -0.114 * 1.772 / 0.587, 1.772],
            [1.402, -0.299 * 1.402 / 0.587, 0.000],
        ],
        dtype=np.float32,
    )

    arr = arr.astype(np.float32)
    arr -= [0, 128, 128]

    # Round(x) -> floor of (arr + 0.5)
    np.matmul(arr, ybr_to_rgb, out=arr)
    arr += 0.5
    np.floor(arr, out=arr)
    # Max(0, arr) -> 0 if 0 >= arr, arr otherwise
    # Min(arr, 255) -> arr if arr <= 255, 255 otherwise
    np.clip(arr, 0, 255, out=arr)

    return arr.astype(orig_dtype)


def create_icc_transform(
    ds: "Dataset | None" = None,
    icc_profile: bytes = b"",
    intent: int | None = None,
    color_space: str | None = None,
) -> "PIL.ImageCms.ImageCmsTransform":
    """Return a Pillow color transformation object from either the dataset `ds` or an
    ICC profile `icc_profile`.

    .. versionadded:: 3.0

    .. warning::

        This function requires `NumPy <https://numpy.org/>`_ and `Pillow
        <https://pillow.readthedocs.io/en/stable/>`_.

    Parameters
    ----------
    ds : pydicom.dataset.Dataset, optional
        Required if `icc_profile` is not supplied, a :class:`~pydicom.dataset.Dataset`
        containing elements from the :dcm:`ICC Profile<part03/sect_C.11.15.html>` module.
    icc_profile : bytes, optional
        Required if `ds` is not supplied, an ICC profile encoded as :class:`bytes`.
    intent : int, optional
        The rendering intent of the transformation, one of:

        * ``0``: perceptual
        * ``1``: relative colorimetric
        * ``2``: saturation
        * ``3``: absolute colorimetric

        If no `intent` is specified then the default rendering intent in the ICC
        profile will be used.
    color_space : str, optional
        The output color space to use for the transformation created from `ds`. If not
        used then defaults to the (0028,2002) *Color Space* value if its present in
        `ds`, otherwise defaults to ``"sRGB"``. If used then it will override the
        (0028,2002) *Color Space* element value (if present). Must be one of the
        :func:`supported Pillow color spaces<PIL.ImageCms.createProfile>`, although in
        practice only ``"sRGB"`` is likely to be used.

    Returns
    -------
    PIL.ImageCms.ImageCmsTransform
        A color transformation object that can be used with :func:`apply_icc_profile`.
    """
    if not HAVE_PIL:
        raise ImportError("Pillow is required to create a color transformation object")

    if ds is None and not icc_profile:
        raise ValueError("Either 'ds' or 'icc_profile' must be supplied")

    if ds and icc_profile:
        raise ValueError("Only one of 'ds' and 'icc_profile' should be used, not both")

    icc_profile = getattr(ds, "ICCProfile", icc_profile)
    if not icc_profile:
        raise ValueError("No (0028,2000) 'ICC Profile' element was found in 'ds'")

    # If ColorSpace is in `ds` try and use that, but allow override via `color_space`
    cs = color_space if color_space else getattr(ds, "ColorSpace", "sRGB")
    if cs and cs.upper() not in _CMS_COLOR_SPACES:
        if not color_space:
            msg = (
                f"The (0028,2002) 'Color Space' value '{cs}' is not supported by "
                "Pillow, please use the 'color_space' argument to specify a "
                "supported value"
            )
        else:
            msg = (
                f"Unsupported 'color_space' value '{cs}', must be 'sRGB', 'LAB' or "
                "'XYZ'"
            )

        raise ValueError(msg)

    # Conform the supplied color space to the value required by Pillow
    cs = _CMS_COLOR_SPACES[cs.upper()]

    profile = ImageCms.ImageCmsProfile(BytesIO(icc_profile))
    intent = intent if intent else ImageCms.getDefaultIntent(profile)
    if intent not in _CMS_INTENTS:
        raise ValueError(f"Invalid 'intent' value '{intent}', must be 0, 1, 2 or 3")

    return ImageCms.buildTransform(
        outputProfile=ImageCms.createProfile(cs),  # type: ignore[arg-type]
        inputProfile=profile,
        inMode="RGB",
        outMode="RGB",
        renderingIntent=ImageCms.Intent(intent),
    )


def _expand_segmented_lut(
    data: tuple[int, ...],
    fmt: str,
    nr_segments: int | None = None,
    last_value: int | None = None,
) -> list[int]:
    """Return a list containing the expanded lookup table data.

    Parameters
    ----------
    data : tuple of int
        The decoded segmented palette lookup table data. May be padded by a
        trailing null.
    fmt : str
        The format of the data, should contain `'B'` for 8-bit, `'H'` for
        16-bit, `'<'` for little endian and `'>'` for big endian.
    nr_segments : int, optional
        Expand at most `nr_segments` from the data. Should be used when
        the opcode is ``2`` (indirect). If used then `last_value` should also
        be used.
    last_value : int, optional
        The previous value in the expanded lookup table. Should be used when
        the opcode is ``2`` (indirect). If used then `nr_segments` should also
        be used.

    Returns
    -------
    list of int
        The reconstructed lookup table data.

    References
    ----------

    * DICOM Standard, Part 3, Annex C.7.9
    """
    # Indirect segment byte offset is dependent on endianness for 8-bit
    # Little endian: e.g. 0x0302 0x0100, big endian, e.g. 0x0203 0x0001
    indirect_ii = [3, 2, 1, 0] if "<" in fmt else [2, 3, 0, 1]

    lut: list[int] = []
    offset = 0
    segments_read = 0
    # Use `offset + 1` to account for possible trailing null
    #   can do this because all segment types are longer than 2
    while offset + 1 < len(data):
        opcode = data[offset]
        length = data[offset + 1]
        offset += 2

        if opcode == 0:
            # C.7.9.2.1: Discrete segment
            lut.extend(data[offset : offset + length])
            offset += length
        elif opcode == 1:
            # C.7.9.2.2: Linear segment
            if lut:
                y0 = lut[-1]
            elif last_value:
                # Indirect segment with linear segment at 0th offset
                y0 = last_value
            else:
                raise ValueError(
                    "Error expanding a segmented palette color lookup table: "
                    "the first segment cannot be a linear segment"
                )

            y1 = data[offset]
            offset += 1

            if y0 == y1:
                lut.extend([y1] * length)
            else:
                step = (y1 - y0) / length
                vals = np.around(np.linspace(y0 + step, y1, length))
                lut.extend([int(vv) for vv in vals])
        elif opcode == 2:
            # C.7.9.2.3: Indirect segment
            if not lut:
                raise ValueError(
                    "Error expanding a segmented palette color lookup table: "
                    "the first segment cannot be an indirect segment"
                )

            if "B" in fmt:
                # 8-bit segment entries
                ii = [data[offset + vv] for vv in indirect_ii]
                byte_offset = (ii[0] << 8 | ii[1]) << 16 | (ii[2] << 8 | ii[3])
                offset += 4
            else:
                # 16-bit segment entries
                byte_offset = data[offset + 1] << 16 | data[offset]
                offset += 2

            lut.extend(_expand_segmented_lut(data[byte_offset:], fmt, length, lut[-1]))
        else:
            raise ValueError(
                "Error expanding a segmented palette lookup table: "
                f"unknown segment type '{opcode}'"
            )

        segments_read += 1
        if segments_read == nr_segments:
            return lut

    return lut
